#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""This module contains the plumbing necessary for creating, modifying and
querying the sharding information from the state stores.

"""

import mysql.fabric.errors as _errors
import mysql.fabric.persistence as _persistence
import mysql.fabric.utils as _utils

from mysql.fabric.server import MySQLServer, Group

class ShardMapping(_persistence.Persistable):
    """Represents the mapping between the sharding scheme and the table
    being sharded. The class encapsulates the operations required to
    manage the information in the state store related to this mapping.

    NOTE: A sharding scheme contains the details of the type name (i.e.) RANGE,
    HASH, List used for sharding and the parameters for the same. For example,
    a range sharding would contain the Lower Bound and the Upper Bound for
    the sharding definition.

    The information in the state store, for a sharding scheme, in the case
    where the state store is a relational store takes the following form:

        +--------------+--------------------------+---------------------+
        | shard_map_id |      table_name          |    column_name      |
        +==============+==========================+=====================+
        |1             |Employee                  |ID                   |
        +--------------+--------------------------+---------------------+
        |2             |Salary                    |EmpID                |
        +--------------+--------------------------+---------------------+

    The columns are explained as follows,
    * shard_mapping_id - The unique identification for a shard mapping.
    * table_name - The tables associated with this shard mapping.
    * column_name - The column name in the table that is used to shard this
    table.

        +--------------+--------------------------+---------------------+
        | shard_map_id |         type_name        |     global          |
        +==============+==========================+=====================+
        |1             |RANGE                     |GROUPIDX             |
        +--------------+--------------------------+---------------------+

    The columns are explained as follows

    * shard_mapping_id - The unique identification for a shard mapping.
    * type_name - The type name of the sharding scheme- RANGE, HASH, LIST etc
    * global_group - Every shard mapping is associated with a Global Group
                        that stores the global updates and the schema changes
                        for this shard mapping and dissipates these to the
                        shards.
    """

    #Create the schema for the tables used to store the shard specification
    #mapping information
    CREATE_SHARD_MAPPING = ("CREATE TABLE shard_tables"
                            "(shard_mapping_id INT NOT NULL, "
                            "table_name VARCHAR(64) NOT NULL, "
                            "column_name VARCHAR(64) NOT NULL, "
                            "PRIMARY KEY (table_name, column_name), "
                            "INDEX(shard_mapping_id))")

    DUMP_SHARD_TABLES = (
                            "SELECT "
                            "table_name, column_name, shard_mapping_id "
                            "FROM "
                            "shard_tables "
                            "WHERE "
                            "shard_mapping_id LIKE %s "
                            "ORDER BY shard_mapping_id, table_name, column_name"
                         )

    DUMP_SHARDING_INFORMATION = (
        "SELECT t.table_name, t.column_name, "
        "IF( m.type_name = 'HASH', HEX(r.lower_bound), r.lower_bound) "
        "AS lower_bound, "
        "r.shard_id, "
        "m.type_name, s.group_id, m.global_group "
        "FROM "
        "shard_maps AS m RIGHT JOIN shard_tables AS t USING (shard_mapping_id) "
        "LEFT JOIN shard_ranges AS r USING (shard_mapping_id) "
        "LEFT JOIN shards AS s USING (shard_id) "
        "WHERE table_name LIKE %s AND "
        "s.state = 'ENABLED' "
        "ORDER BY r.shard_id, t.table_name, t.column_name, r.lower_bound"
    )

    CREATE_SHARD_MAPPING_DEFN = ("CREATE TABLE shard_maps"
                                 "(shard_mapping_id INT AUTO_INCREMENT "
                                 "NOT NULL PRIMARY KEY, "
                                 "type_name ENUM('RANGE', 'HASH') NOT NULL, "
                                 "global_group VARCHAR(64))")

    DUMP_SHARD_MAPS = (
                            "SELECT "
                            "shard_mapping_id, type_name, global_group "
                            "FROM "
                            "shard_maps "
                            "WHERE "
                            "shard_mapping_id LIKE %s "
                            "ORDER BY "
                            "shard_mapping_id, type_name, global_group"
                       )

    #Create the referential integrity constraint with the shard_maps
    #table.
    ADD_FOREIGN_KEY_CONSTRAINT_SHARD_MAPPING_ID = \
                                ("ALTER TABLE shard_tables "
                                  "ADD CONSTRAINT fk_shard_mapping_id "
                                  "FOREIGN KEY(shard_mapping_id) REFERENCES "
                                  "shard_maps(shard_mapping_id)")

    #Create the referential integrity constraint with the groups table.
    ADD_FOREIGN_KEY_CONSTRAINT_GLOBAL_GROUP = \
                            ("ALTER TABLE shard_maps "
                            "ADD CONSTRAINT fk_shard_mapping_global_group "
                            "FOREIGN KEY(global_group) REFERENCES "
                            "groups(group_id)")

    #Define a shard mapping.
    DEFINE_SHARD_MAPPING = ("INSERT INTO "
                            "shard_maps(type_name, global_group) "
                            "VALUES(%s, %s)")
    ADD_SHARD_MAPPING = ("INSERT INTO shard_tables"
                         "(shard_mapping_id, table_name, column_name) "
                         "VALUES(%s, %s, %s)")

    #select the shard specification mapping information for a table
    SELECT_SHARD_MAPPING = ("SELECT sm.shard_mapping_id, table_name, "
                            "column_name, type_name, "
                            "global_group "
                            "FROM shard_tables as sm, "
                            "shard_maps as smd "
                            "WHERE sm.shard_mapping_id = smd.shard_mapping_id "
                            "AND table_name = %s")

    #Select the shard mapping for a given shard mapping ID.
    SELECT_SHARD_MAPPING_BY_ID = ("SELECT sm.shard_mapping_id, table_name, "
                            "column_name, type_name, "
                            "global_group "
                            "FROM shard_tables as sm, "
                            "shard_maps as smd "
                            "WHERE sm.shard_mapping_id = smd.shard_mapping_id "
                            "AND sm.shard_mapping_id = %s")

    #Select all the shard specifications of a particular sharding type name.
    LIST_SHARD_MAPPINGS = ("SELECT sm.shard_mapping_id, table_name, "
                            "column_name, type_name, "
                            "global_group "
                            "FROM shard_tables as sm, "
                            "shard_maps as smd "
                            "WHERE sm.shard_mapping_id = smd.shard_mapping_id "
                            "AND type_name = %s")

    #Select the shard mapping for a given shard mapping ID.
    SELECT_SHARD_MAPPING_DEFN = ("SELECT shard_mapping_id, type_name, "
                                 "global_group FROM shard_maps "
                                 "WHERE shard_mapping_id = %s")

    #Select all the shard mapping definitions
    LIST_SHARD_MAPPING_DEFN = ("SELECT shard_mapping_id, type_name, "
                               "global_group FROM shard_maps")

    #Delete the sharding specification to table mapping for a given table.
    DELETE_SHARD_MAPPING = ("DELETE FROM shard_tables "
                            "WHERE table_name = %s")

    #Delete the shard mapping definition
    DELETE_SHARD_MAPPING_DEFN = ("DELETE FROM shard_maps "
                                 "WHERE shard_mapping_id = %s")

    QUERY_SHARD_MAPPING_PER_GROUP = ("SELECT shard_mapping_id "
        "FROM shard_maps WHERE global_group = %s"
    )

    def __init__(self, shard_mapping_id, table_name, column_name, type_name,
                 global_group):
        """Initialize the Shard Specification Mapping for a given table.

        :param shard_mapping_id: The unique identifier for this shard mapping.
        :param table_name: The table for which the mapping is being created.
        :param column_name: The column name on which the mapping is being
                            defined. The column is useful when resharding
                            comes into the picture.
        :param type_name: The type of sharding defined on the given table.
                            E.g. RANGE, HASH etc.
        :param global_group: Every shard mapping is associated with a Global
                                Group that stores the global updates and the
                                schema changes for this shard mapping and
                                dissipates these to the shards.
        """
        super(ShardMapping, self).__init__()
        self.__shard_mapping_id = shard_mapping_id
        self.__table_name = table_name
        self.__column_name = column_name
        self.__type_name = type_name
        self.__global_group = global_group

    @property
    def shard_mapping_id(self):
        """Return the shard mapping id for this shard mapping.
        """
        return self.__shard_mapping_id

    @property
    def table_name(self):
        """Return the table on which the shard mapping is defined.
        """
        return self.__table_name

    @property
    def column_name(self):
        """Return the column in the table on which the sharding is defined.
        """
        return self.__column_name

    @property
    def type_name(self):
        """Return the type of the sharding specification defined on the table.
        """
        return self.__type_name

    @property
    def global_group(self):
        """Return the global group for the sharding specification.
        """
        return self.__global_group

    @staticmethod
    def list(sharding_type, persister=None):
        """The method returns all the shard mappings (names) of a
        particular sharding_type. For example if the method is called
        with 'range' it returns all the sharding specifications that exist
        of type range.

        :param sharding_type: The sharding type for which the sharding
                              specification needs to be returned.
        :param persister: A valid handle to the state store.

        :return: A list of sharding specification names that are of the
                  sharding type.
        """

        sharding_type = sharding_type.upper()
        cur = persister.exec_stmt(ShardMapping.LIST_SHARD_MAPPINGS,
                                  {"raw" : False,
                                  "fetch" : False,
                                  "params" : (sharding_type,)})
        rows = cur.fetchall()

        return [ ShardMapping(*row[0:5]) for row in rows ]

    def remove(self, persister=None):
        """Remove the shard mapping represented by the Shard Mapping object.
        The method basically results in removing the association between a
        table and the sharding specification used to shard this table.

        :param persister: Represents a valid handle to the state
                          store.
        """
        #Remove the shard mapping
        persister.exec_stmt(
            ShardMapping.DELETE_SHARD_MAPPING,
            {"params":(self.__table_name,)})

    @staticmethod
    def remove_sharding_definition(shard_mapping_id, persister=None):
        """Remove the shard mapping definition identified uniquely by the given
        shard_mapping_id.

        :param shard_mapping_id: The shard mapping ID of the sharding defn.
        :param persister: Represents a valid handle to the state
                          store.
        """
        #Remove the shard mapping definition.
        persister.exec_stmt(
            ShardMapping.DELETE_SHARD_MAPPING_DEFN,
            {"params":(shard_mapping_id,)})

    @staticmethod
    def create(persister=None):
        """Create the schema to store the mapping between the table and
        the sharding specification.

        :param persister: A valid handle to the state store.
        """

        persister.exec_stmt(ShardMapping.CREATE_SHARD_MAPPING)
        persister.exec_stmt(ShardMapping.CREATE_SHARD_MAPPING_DEFN)

    @staticmethod
    def fetch(table_name, persister=None):
        """Fetch the shard specification mapping for the given table

        :param table_name: The name of the table for which the sharding
                            specification is being queried.
        :param persister: A valid handle to the state store.
        :return: The ShardMapping object that encapsulates the shard mapping
                    information for the given table.
        """
        cur = persister.exec_stmt(
                                  ShardMapping.SELECT_SHARD_MAPPING,
                                  {"raw" : False,
                                  "fetch" : False,
                                  "params" : (table_name,)})
        row = cur.fetchone()
        if row:
            return ShardMapping(row[0], row[1], row[2], row[3], row[4])

        return None

    @staticmethod
    def fetch_by_id(shard_mapping_id, persister=None):
        """Fetch the shard specification mapping for the given shard mapping
        ID.

        :param shard_mapping_id: The shard mapping id for which the sharding
                            specification is being queried.
        :param persister: A valid handle to the state store.
        :return: The ShardMapping objects that encapsulates the shard mapping
                    information for the given shard mapping ID.
        """
        #When there are several tables related (sharded) by the same key defn,
        #we will have multiple shard mappings for the same shard mapping key.
        shard_mapping_list = []

        cur = persister.exec_stmt(
                                  ShardMapping.SELECT_SHARD_MAPPING_BY_ID,
                                  {"raw" : False,
                                  "fetch" : False,
                                  "params" : (shard_mapping_id,)})
        rows = cur.fetchall()

        for row in rows:
            shard_mapping_list.append(
                ShardMapping(row[0], row[1], row[2], row[3], row[4])
            )

        return shard_mapping_list

    @staticmethod
    def fetch_shard_mapping_defn(shard_mapping_id, persister=None):
        """Fetch the shard mapping definition corresponding to the
        shard mapping id.

        :param shard_mapping_id: The id of the shard mapping definition that
                                    needs to be fetched.
        :param persister: A valid handle to the state store.
        :return: A list containing the shard mapping definition parameters.
        """
        row = persister.exec_stmt(ShardMapping.SELECT_SHARD_MAPPING_DEFN,
                                  {"raw" : False,
                                  "fetch" : True,
                                  "params" : (shard_mapping_id,)})
        if row:
            #There is no abstraction for a shard mapping definition. A
            #shard mapping definition is just a triplet of
            #(shard_id, sharding_type, global_group)
            return row[0]

        return None

    @staticmethod
    def list_shard_mapping_defn(persister=None):
        """Fetch all the shard mapping definitions.

        :param persister: A valid handle to the state store.
        :return: A list containing the shard mapping definitions.
        """
        rows = persister.exec_stmt(ShardMapping.LIST_SHARD_MAPPING_DEFN,
                                  {"raw" : False,
                                  "fetch" : True})
        if rows is not None:
            return rows

        return []

    @staticmethod
    def lookup_shard_mapping_id(group_id, persister=None):
      """Return the shard mapping associated with a group.

      :group_id: Group identification.
      :param persister: A valid handle to the state store.
      :return: Shard mapping associated to a group..
      """
      rows = persister.exec_stmt(ShardMapping.QUERY_SHARD_MAPPING_PER_GROUP, 
          {"params": (group_id, ), "raw" : False, "fetch" : True}
      )
      if rows:
          return rows[0][0]

    @staticmethod
    def define(type_name, global_group_id, persister=None):
        """Define a shard mapping.

        :param type_name: The type of sharding scheme - RANGE, HASH, LIST etc
        :param global_group: Every shard mapping is associated with a
                            Global Group that stores the global updates
                            and the schema changes for this shard mapping
                            and dissipates these to the shards.
        :param persister: A valid handle to the state store.
        :return: The shard_mapping_id generated for the shard mapping.
        """
        persister.exec_stmt(
            ShardMapping.DEFINE_SHARD_MAPPING,
            {"params":(type_name, global_group_id)})
        row = persister.exec_stmt("SELECT LAST_INSERT_ID()")
        return int(row[0][0])

    @staticmethod
    def add(shard_mapping_id, table_name, column_name, persister=None):
        """Add a table to a shard mapping.

        :param shard_mapping_id: The shard mapping id to which the input
                                    table is attached.
        :param table_name: The table being sharded.
        :param column_name: The column whose value is used in the sharding
                            scheme being applied
        :param persister: A valid handle to the state store.

        :return: The ShardMapping object for the mapping created.
                 None if the insert failed.
        """
        persister.exec_stmt(
            ShardMapping.ADD_SHARD_MAPPING,
            {"params":(shard_mapping_id, table_name, column_name)})
        return ShardMapping.fetch(table_name)

    @staticmethod
    def dump_shard_tables(version=None, patterns="", persister=None):
        """Return the list of shard tables that belong to the shard mappings
        listed in the patterns strings separated by comma.

        :param version: The connectors version of the data.
        :param patterns: shard mapping pattern.
        :param persister: Persister to persist the object to.
        :return: A list of registered tables using the provided shard mapping
                pattern.
        """
        #This is used to store the consolidated list of shard mappings
        #that will be returned.
        result_shard_tables_list = []

        #This stores the pattern that will be passed to the LIKE MySQL
        #command.
        like_pattern = None

        if patterns is None:
            patterns = ''

        #Split the patterns string into a list of patterns of groups.
        pattern_list = _utils.split_dump_pattern(patterns)

        #Iterate through the pattern list and fire a query for
        #each pattern.
        for find in pattern_list:
            if find == '':
                like_pattern = '%%'
            else:
                like_pattern = '%' + find + '%'
            cur = persister.exec_stmt(ShardMapping.DUMP_SHARD_TABLES,
                                  {"fetch" : False, "params":(like_pattern,)})
            rows = cur.fetchall()
            #For each row fetched, split the fully qualified table name into
            #a database and table name.
            for row in rows:
                database, table = _utils.split_database_table(row[0])
                result_shard_tables_list.append(
                    (
                        database,
                        table,
                        row[1],
                        row[2]
                    )
                )
        return _utils.wrap_output(result_shard_tables_list)

    @staticmethod
    def dump_sharding_info(version=None, patterns="", persister=None):
        """Return all the sharding information about the table in the
        patterns string.

        :param version: The connectors version of the data.
        :param patterns: shard table pattern.
        :param persister: Persister to persist the object to.
        :return: The sharding information for all the tables passed in patterns.
        """
        #This is used to store the consolidated list of shard mappings
        #that will be returned.
        result_shard_tables_list = []

        #This stores the pattern that will be passed to the LIKE MySQL
        #command.
        like_pattern = None

        if patterns is None:
            patterns = ''

        #Split the patterns string into a list of patterns of groups.
        pattern_list = _utils.split_dump_pattern(patterns)

        #Iterate through the pattern list and fire a query for
        #each pattern.
        for find in pattern_list:
            if not find:
                like_pattern = '%%'
            else:
                like_pattern = find
            cur = persister.exec_stmt(ShardMapping.DUMP_SHARDING_INFORMATION,
                                  {"fetch" : False, "params":(like_pattern,)})
            rows = cur.fetchall()
            #For each row fetched, split the fully qualified table name into
            #a database and table name.
            for row in rows:
                database, table = _utils.split_database_table(row[0])
                result_shard_tables_list.append(
                    (
                        database,
                        table,
                        row[1],  # column_name
                        row[2],  # lower_bound
                        row[3],  # shard_id
                        row[4],  # group_id
                        row[5],  # global_group
                        row[6],  # type
                    )
                )
        return _utils.wrap_output(result_shard_tables_list)

    @staticmethod
    def dump_shard_maps(version=None, patterns="", persister=None):
        """Return the list of shard mappings identified by the shard mapping
        IDs listed in the patterns strings separated by comma.

        :param version: The connectors version of the data.
        :param patterns: shard mapping pattern.
        :param persister: Persister to persist the object to.
        :return: A list of shard mappings matching the provided pattern.
        """
        #This is used to store the consolidated list of shard mappings
        #that will be returned.
        result_shard_maps = []

        #This stores the pattern that will be passed to the LIKE MySQL
        #command.
        like_pattern = None

        if patterns is None:
            patterns = ''

        #Split the patterns string into a list of patterns of groups.
        pattern_list = _utils.split_dump_pattern(patterns)

        #Iterate through the pattern list and fire a query for
        #each pattern.
        for find in pattern_list:
            if find == '':
                like_pattern = '%%'
            else:
                like_pattern = '%' + find + '%'
            cur = persister.exec_stmt(ShardMapping.DUMP_SHARD_MAPS,
                                  {"fetch" : False, "params":(like_pattern,)})
            rows = cur.fetchall()
            for row in rows:
                result_shard_maps.append((row[0], row[1], row[2]))
        return _utils.wrap_output(result_shard_maps)

class Shards(_persistence.Persistable):
    """Contains the mapping between the Shard ID and the Group ID

    A typical mapping between the shards and their location looks like
    the following

+--------------------+--------------------+--------------------+
|     shard_id       |      group_id      |     state          |
+====================+====================+====================+
|1                   |GroupID1            |ENABLED             |
+--------------------+--------------------+--------------------+

    The columns are explained as follows,

    shard_id - Unique identifier for the shard of a particular table.
    group_id - The Server location for the given partition of the table.
    state - Indicates whether a given shard is ENABLED or DISABLED.
    """

    #Tuple stores the list of valid shard mapping types.
    VALID_SHARDING_TYPES = ('RANGE', 'HASH')

    #Valid states of the shards
    VALID_SHARD_STATES = ("ENABLED", "DISABLED")

    #Create the schema for storing the shard to groups mapping
    CREATE_SHARDS = ("CREATE TABLE shards ("
                    "shard_id INT AUTO_INCREMENT NOT NULL PRIMARY KEY, "
                    "group_id VARCHAR(64) NOT NULL, "
                    "state ENUM('DISABLED', 'ENABLED') NOT NULL)")

    #Create the referential integrity constraint with the groups table.
    ADD_FOREIGN_KEY_CONSTRAINT_GROUP_ID = \
                                ("ALTER TABLE shards "
                                  "ADD CONSTRAINT fk_shards_group_id "
                                  "FOREIGN KEY(group_id) REFERENCES "
                                  "groups(group_id)")

    #Insert the Range to Shard mapping into the table.
    INSERT_SHARD = ("INSERT INTO shards(group_id, state) VALUES(%s, %s)")

    #Update the group_id for a shard.
    UPDATE_SHARD = ("UPDATE shards SET group_id=%s WHERE shard_id=%s")

    #Delete a given shard to group mapping.
    DELETE_SHARD = ("DELETE FROM shards WHERE shard_id = %s")

    #Select the group to which a shard ID maps to.
    SELECT_SHARD = ("SELECT shard_id, group_id, state "
                                    "FROM shards WHERE shard_id = %s")

    #Dump all the shard indexes that belong to a shard mapping ID.
    DUMP_SHARD_INDEXES = (
                            "SELECT "
                            "IF( m.type_name = 'HASH', HEX(sr.lower_bound), "
                            "sr.lower_bound), sr.shard_mapping_id, "
                            "s.shard_id, s.group_id "
                            "FROM "
                            "shard_maps AS m JOIN shard_ranges AS sr "
                            "USING (shard_mapping_id) "
                            "JOIN shards AS s USING (shard_id) "
                            "WHERE sr.shard_mapping_id LIKE %s AND "
                            "s.state = 'ENABLED' "
                            "ORDER BY s.shard_id, sr.shard_mapping_id, "
                            "sr.lower_bound, s.group_id"
                            )

    #Select the shard that belongs to a given group.
    SELECT_GROUP_FOR_SHARD = ("SELECT shard_id FROM shards WHERE group_id = %s")

    #Update the state of a shard
    UPDATE_SHARD_STATE = ("UPDATE shards SET state=%s where shard_id=%s")

    def __init__(self, shard_id, group_id,  state="DISABLED"):
        """Initialize the Shards object with the shard to group mapping.

        :param shard_id: An unique identification, a logical representation for
                         a shard of a particular table.
        :param group_id: The group ID to which the shard maps to.
        :param state: Indicates whether a given shard is ENABLED or DISABLED
        """
        super(Shards, self).__init__()
        self.__shard_id = shard_id
        self.__group_id = group_id
        self.__state = state

    @staticmethod
    def create(persister=None):
        """Create the schema to store the current Shard to Group mapping.

        :param persister: A valid handle to the state store.
        """
        persister.exec_stmt(Shards.CREATE_SHARDS)

    @staticmethod
    def add(group_id, state="DISABLED", persister=None):
        """Add a Group that will store a shard. A shard ID is automatically
        generated for a given added Group.

        :param group_id: The Group that is being added to store a shard.
        :param persister: A valid handle to the state store.
        :param state: Indicates whether a given shard is ENABLED or DISABLED

        :return: The Shards object containing a mapping between the shard
                    and the group.
        """
        persister.exec_stmt(Shards.INSERT_SHARD, {"params":(group_id, state)})
        row = persister.exec_stmt("SELECT LAST_INSERT_ID()")
        return Shards(int(row[0][0]), group_id, state)

    @staticmethod
    def add_constraints(persister=None):
        """Add the constraints on the Shards tables.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(Shards.ADD_FOREIGN_KEY_CONSTRAINT_GROUP_ID)

    def remove(self, persister=None):
        """Remove the Shard to Group mapping.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(Shards.DELETE_SHARD, \
                            {"params":(self.__shard_id,)})

    @staticmethod
    def fetch(shard_id, persister=None):
        """Fetch the Shards object containing the group_id for the input
        shard ID.

        :param shard_id: That shard ID whose mapping needs to be fetched.
        :param persister: The DB server that can be used to access the
                          state store.
        """
        row = persister.exec_stmt(Shards.SELECT_SHARD, \
                                  {"params":(shard_id,)})

        if row is None:
            return None

        return Shards(row[0][0], row[0][1], row[0][2],)

    def enable(self, persister=None):
        """Set the state of the shard to ENABLED.
        """
        persister.exec_stmt(
          Shards.UPDATE_SHARD_STATE,
                             {"params":('ENABLED', self.__shard_id)})

    def disable(self, persister=None):
        """Set the state of the shard to DISABLED.
        """
        persister.exec_stmt(
          Shards.UPDATE_SHARD_STATE,
                             {"params":('DISABLED', self.__shard_id)})

    @staticmethod
    def lookup_shard_id(group_id,  persister=None):
        """Fetch the shard ID for the given Group.

        :param group_id: The Group that is being looked up.
        :param persister: A valid handle to the state store.

        :return: The shard_id contained in the given group_id.
        """
        row = persister.exec_stmt(Shards.SELECT_GROUP_FOR_SHARD, \
                                  {"params":(group_id,)})
        if row:
            return row[0][0]

    @property
    def shard_id(self):
        """Return the shard ID for the Shard to Group mapping.
        """
        return self.__shard_id

    @property
    def group_id(self):
        """Return the Group ID for the Shard to Group mapping.
        """
        return self.__group_id

    @group_id.setter
    def group_id(self,  group_id,  persister=None):
        """Set the group_id for the Shard.

        :param group_id: The Group that is being added to store a shard.
        :param persister: A valid handle to the state store.
        """
        persister.exec_stmt(Shards.UPDATE_SHARD,
                                        {"params":(group_id, self.__shard_id)})
        self.__group_id = group_id

    @property
    def state(self):
        """Return whether the state is ENABLED or DISABLED.
        """
        return self.__state

    @staticmethod
    def dump_shard_indexes(version=None, patterns="", persister=None):
        """Return the list of shard indexes that belong to the shard mappings
        listed in the patterns strings separated by comma.

        :param version: The connectors version of the data.
        :param patterns: shard mapping pattern.
        :param persister: Persister to persist the object to.
        :return: A list of shard indexes belonging to the shard mappings that
                match the provded patterns.
        """
        #This is used to store the consolidated list of shard mappings
        #that will be returned.
        result_shard_indexes_list = []

        #This stores the pattern that will be passed to the LIKE MySQL
        #command.
        like_pattern = None

        if patterns is None:
            patterns = ''

        #Split the patterns string into a list of patterns of groups.
        pattern_list = _utils.split_dump_pattern(patterns)

        #Iterate through the pattern list and fire a query for
        #each pattern.
        for find in pattern_list:
            if find == '':
                like_pattern = '%%'
            else:
                like_pattern = '%' + find + '%'
            cur = persister.exec_stmt(Shards.DUMP_SHARD_INDEXES,
                                  {"fetch" : False, "params":(like_pattern,)})
            rows = cur.fetchall()
            for row in rows:
                result_shard_indexes_list.append(
                    (
                        row[0], row[1], row[2], row[3]
                    )
                )
        return _utils.wrap_output(result_shard_indexes_list)

class RangeShardingIntegerHandler(object):
    """Contains the members that are required to handle a sharding definition
    based on an INTEGER datatype.
    """

   #Select the server corresponding to the RANGE to which a given key
    #belongs. The query either selects the least lower_bound that is larger
    #than a given key or selects the largest lower_bound and insert the key
    #in that shard.
    LOOKUP_KEY = (
        "SELECT "
        "sr.shard_mapping_id, "
        "sr.lower_bound, "
        "s.shard_id "
        "FROM "
        "shard_ranges AS sr, shards AS s "
        "WHERE %s >= CAST(lower_bound AS SIGNED) "
        "AND sr.shard_mapping_id = %s "
        "AND s.shard_id = sr.shard_id "
        "ORDER BY CAST(lower_bound AS SIGNED) DESC "
        "LIMIT 1"
    )

    #Select the UPPER BOUND for a given LOWER BOUND value.
    SELECT_UPPER_BOUND = (
        "SELECT lower_bound FROM "
        "shard_ranges "
        "WHERE  CAST(lower_bound AS SIGNED) > %s AND shard_mapping_id = %s "
        "ORDER BY  CAST(lower_bound AS SIGNED) ASC LIMIT 1"
    )

    @staticmethod
    def is_valid_lower_bound(lower_bound):
        """Verify if the given value is a valid INTEGER lower bound.

        :param lower_bound: The value that needs to be verified.
        """
        try:
            int(lower_bound)
            return True
        except ValueError:
            return False

    @staticmethod
    def split_value(lower_bound, upper_bound):
        """Find the middle value of a sharding range that can be used when
        the split value is not explicitly specified.

        :param lower_bound: The lower bound of the shard range.
        :param upper_bound: The upper bound of the shard range.
        """
        #If the underlying sharding specification is a RANGE, and the
        #split value is not given, then calculate it as the mid value
        #between the current lower_bound and its next lower_bound.
        lower_bound = int(lower_bound)
        upper_bound = int(upper_bound)
        split_value = lower_bound + (upper_bound - lower_bound) / 2
        return split_value

    @staticmethod
    def is_valid_split_value(split_value, lower_bound, upper_bound):
        """Verify if the given split value for a RANGE sharding definition is
        valid.

        :param split_value: The split value.
        :param lower_bound: The lower bound of the sharding definition.
        :param upper_bound: The upper bound of the sharding definition.

        :return: True - If the split value is valid
                False - If the split value is not valid
        """
        #A split value needs to be a valid lower_bound value and should
        #lie between the lower bound and the upper bound values
        if not RangeShardingIntegerHandler.is_valid_lower_bound(split_value):
            return False
        if (int(split_value) > int(lower_bound)) is not True:
            return False
        if upper_bound is not None:
            if (int(split_value) < int(upper_bound)) is not True:
                return False
        return True

class HashShardingIntegerHandler(object):
    """Contains the members that are required to handle a hash based
    sharding definition.
    """
    #Fetch the shard ID corresponding to the input key.
    LOOKUP_KEY = (
                "("
                "SELECT "
                "sr.shard_mapping_id, "
                "HEX(sr.lower_bound) AS lower_bound, "
                "s.shard_id "
                "FROM shard_ranges AS sr, shards AS s "
                "WHERE MD5(%s) >= HEX(sr.lower_bound) "
                "AND sr.shard_mapping_id = %s "
                "AND s.shard_id = sr.shard_id "
                "ORDER BY HEX(sr.lower_bound) DESC "
                "LIMIT 1"
                ") "
                "UNION ALL "
                "("
                "SELECT "
                "sr.shard_mapping_id, "
                "HEX(sr.lower_bound) AS lower_bound, "
                "sr.shard_id "
                "FROM shard_ranges AS sr, shards AS s "
                "WHERE sr.shard_mapping_id = %s "
                "AND s.shard_id = sr.shard_id "
                "ORDER BY HEX(sr.lower_bound) DESC "
                "LIMIT 1"
                ") "
                "ORDER BY HEX(lower_bound) ASC "
                "LIMIT 1"
                )

    #Select the UPPER BOUND for a given LOWER BOUND value.
    SELECT_UPPER_BOUND = (
        "SELECT HEX(lower_bound) FROM "
        "shard_ranges "
        "WHERE "
        "HEX(lower_bound) > %s "
        "AND "
        "shard_mapping_id = %s "
        "ORDER BY HEX(lower_bound) ASC LIMIT 1"
    )

    @staticmethod
    def is_valid_lower_bound(lower_bound):
        """Lower bounds in hash based sharding are autogenerated.
        This method is thus a no-op for hash based sharding.
        """
        return False

    @staticmethod
    def split_value(lower_bound, upper_bound):
        """Calculate the split value for a hash based shard as the mid-point
        of the lower and upper bounds of the current shard.

        :param lower_bound: The lower bound of the hash based shard.
        :param upper_bound: The upper bound of the hash based shard.

        :return: Hexadecimal representatin of the mid values.
        """
        #Retrieve an integer representation of the hexadecimal
        #lower_bound. The value is actually a long. Python automatically
        #returns a long value.
        lower_bound = int(lower_bound,  16)
        upper_bound = int(upper_bound, 16)
        #split value after the below computation is actually a long.
        split_value = lower_bound + (upper_bound - lower_bound) / 2
        #split_value after the hex computation gets stored with a prefix
        #0x indicating a hexadecimal value and a suffix of L indicating a
        #Long. Extract the hexadecimal string from this value.
        split_value = "%x" % (split_value)
        return split_value

    @staticmethod
    def is_valid_split_value(split_value, lower_bound, upper_bound):
        return True

class RangeShardingSpecification(_persistence.Persistable):
    """Represents a RANGE sharding specification. The class helps encapsulate
    the representation of a typical RANGE sharding implementation in the
    state store.

    A typical RANGE sharding representation looks like the following,

        +--------------+---------+-----------+
        | shard_map_id |   LB    |  shard_id |
        +==============+=========+===========+
        |1             |10000    |1          |
        +--------------+---------+-----------+

    The columns in the above table are explained as follows,

    * shard_mapping_id - The unique identification for a shard mapping.
    * LB -The lower bound of the given RANGE sharding scheme instance
    * shard_id - An unique identification, a logical representation for a
                    shard of a particular table.
    """

    #Create the schema for storing the RANGE sharding specificaton.
    CREATE_RANGE_SPECIFICATION = ("CREATE TABLE "
                                "shard_ranges "
                                "(shard_mapping_id INT NOT NULL, "
                                "lower_bound VARBINARY(16) NOT NULL, "
                                "INDEX(lower_bound), "
                                "UNIQUE(shard_mapping_id, lower_bound), "
                                "shard_id INT NOT NULL)")

    #Create the referential integrity constraint with the shard_mapping_defn
    #table
    ADD_FOREIGN_KEY_CONSTRAINT_SHARD_MAPPING_ID = (
        "ALTER TABLE shard_ranges "
        "ADD CONSTRAINT "
        "fk_shard_mapping_id_sharding_spec "
        "FOREIGN KEY(shard_mapping_id) REFERENCES "
        "shard_maps(shard_mapping_id)"
    )

    #Create the referential integrity constraint with the shard_id
    #table
    ADD_FOREIGN_KEY_CONSTRAINT_SHARD_ID = \
                                ("ALTER TABLE shard_ranges "
                                  "ADD CONSTRAINT fk_shard_id_sharding_spec "
                                  "FOREIGN KEY(shard_id) REFERENCES "
                                  "shards(shard_id)")

    #Insert a RANGE of keys and the server to which they belong.
    INSERT_RANGE_SPECIFICATION = ("INSERT INTO shard_ranges"
        "(shard_mapping_id, lower_bound, shard_id) VALUES(%s, %s, %s)")

    #Delete a given RANGE specification instance.
    DELETE_RANGE_SPECIFICATION = ("DELETE FROM shard_ranges "
                                  "WHERE "
                                  "shard_id = %s")

    #Given a Shard ID select the RANGE Scheme that it defines.
    SELECT_RANGE_SPECIFICATION = (
        "SELECT shard_mapping_id, lower_bound, "
        "shard_id "
        "FROM shard_ranges "
        "WHERE shard_id = %s"
    )

    #Given a Shard Mapping ID select all the RANGE mappings that it
    #defines.
    LIST_RANGE_SPECIFICATION = (
        "SELECT "
        "shard_mapping_id, "
        "lower_bound, "
        "shard_id "
        "FROM shard_ranges "
        "WHERE shard_mapping_id = %s"
    )

    #Update the Range for a particular shard. The updation needs to happen
    #for the upper bound and the lower bound simultaneously.
    UPDATE_RANGE = (
        "UPDATE shard_ranges SET lower_bound = %s "
        " WHERE shard_id = %s"
    )

    def __init__(self, shard_mapping_id, lower_bound, shard_id):
        """Initialize a given RANGE sharding mapping specification.

        :param shard_mapping_id: The unique identification for a shard mapping.
        :param lower_bound: The lower bound of the given RANGE sharding defn
        :param shard_id: An unique identification, a logical representation
                        for a shard of a particular table.
        """
        super(RangeShardingSpecification, self).__init__()
        self.__shard_mapping_id = shard_mapping_id
        self.__lower_bound = lower_bound
        self.__shard_id = shard_id

    @property
    def shard_mapping_id(self):
        """Return the shard mapping to which this RANGE definition belongs.
        """
        return self.__shard_mapping_id

    @property
    def lower_bound(self):
        """Return the lower bound of this RANGE specification.
        """
        return self.__lower_bound

    @property
    def shard_id(self):
        """Return the Shard ID of the shard for this RANGE sharding
        definition
        """
        return self.__shard_id

    def remove(self, persister=None):
        """Remove the RANGE specification mapping represented by the current
        RANGE shard specification object.

        :param persister: Represents a valid handle to the state store.
        """
        persister.exec_stmt(
            RangeShardingSpecification.DELETE_RANGE_SPECIFICATION,
            {"params":(self.__shard_id,)})

    @staticmethod
    def add(shard_mapping_id, lower_bound, shard_id, persister=None):
        """Add the RANGE shard specification. This represents a single instance
        of a shard specification that maps a key RANGE to a server.

        :param shard_mapping_id: The unique identification for a shard mapping.
        :param lower_bound: The lower bound of the given RANGE sharding defn
        :param shard_id: An unique identification, a logical representation
                        for a shard of a particular table.

        :return: A RangeShardSpecification object representing the current
                Range specification.
                None if the insert into the state store failed
        """
        persister.exec_stmt(
            RangeShardingSpecification.INSERT_RANGE_SPECIFICATION, {
                "params":(
                    shard_mapping_id,
                    lower_bound,
                    shard_id
                )
            }
        )
        return RangeShardingSpecification(
            shard_mapping_id,
            lower_bound,
            shard_id
        )

    @staticmethod
    def create(persister=None):
        """Create the schema to store the current RANGE sharding specification.

        :param persister: A valid handle to the state store.
        """

        persister.exec_stmt(
                    RangeShardingSpecification.CREATE_RANGE_SPECIFICATION)

    @staticmethod
    def list(shard_mapping_id, persister=None):
        """Return the RangeShardingSpecification objects corresponding to the
        given sharding scheme.

        :param shard_mapping_id: The unique identification for a shard mapping.
        :param persister: A valid handle to the state store.

        :return: A  list of RangeShardingSpecification objects which belong to
                to this shard mapping.
                None if the shard mapping is not found
        """

        cur = persister.exec_stmt(
                    RangeShardingSpecification.LIST_RANGE_SPECIFICATION,
                        {"raw" : False,
                        "fetch" : False,
                        "params" : (shard_mapping_id,)})
        rows = cur.fetchall()
        return [ RangeShardingSpecification(*row[0:5]) for row in rows ]

    @staticmethod
    def add_constraints(persister=None):
        """Add the constraints on the sharding tables.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(
            RangeShardingSpecification.ADD_FOREIGN_KEY_CONSTRAINT_SHARD_MAPPING_ID
            )
        persister.exec_stmt(
            RangeShardingSpecification.ADD_FOREIGN_KEY_CONSTRAINT_SHARD_ID
            )

    @staticmethod
    def fetch(shard_id, persister=None):
        """Return the RangeShardingSpecification object corresponding to the
        given sharding ID.

        :param shard_id: The unique identification for a shard.
        :param persister: A valid handle to the state store.

        :return: A list that represents the information in the
                    RangeShardingSpecification.
                    An Empty list otherwise.
        """
        cur = persister.exec_stmt(
                    RangeShardingSpecification.SELECT_RANGE_SPECIFICATION,
                        {"raw" : False,
                        "fetch" : False,
                        "params" : (shard_id,)})
        row = cur.fetchone()
        if row is None:
            return None
        return RangeShardingSpecification(row[0], row[1],  row[2])

    @staticmethod
    def update_shard(shard_id, lower_bound, persister=None):
        """Update the range for a given shard_id.

        :param shard_id: The ID of the shard whose range needs to be updated.
        :param lower_bound: The new lower bound for the shard.
        :param persister: A valid handle to the state store.
        """
        persister.exec_stmt(
            RangeShardingSpecification.UPDATE_RANGE,
            {"params" : (lower_bound, shard_id)}
        )

    @staticmethod
    def lookup(key, shard_mapping_id, type, persister=None):
        """Return the Range sharding specification in whose key range the input
            key falls.

        :param key: The key which needs to be checked to see which range it
                    falls into
        :param shard_mapping_id: The unique identification for a shard mapping.
        :param persister: A valid handle to the state store.

        :return: The Range Sharding Specification that contains the range in
                which the key belongs.
        """
        cur = persister.exec_stmt(SHARDING_DATATYPE_HANDLER[type].LOOKUP_KEY,
                    {"raw" : False,
                    "fetch" : False,
                    "params" : (key, shard_mapping_id)})

        row = cur.fetchone()

        if row is None:
            return None
        return RangeShardingSpecification(row[0], row[1],  row[2])

    @staticmethod
    def get_upper_bound(lower_bound, shard_mapping_id, type, persister=None):
        """Return the next value in range for a given lower_bound value.
        This basically helps to form a (lower_bound, upper_bound) pair
        that can be used during a prune.

        :param lower_bound: The lower_bound value whose next range needs to
                            be retrieved.
        :param shard_mapping_id: The shard_mapping_id whose shards should be
                                searched for the given lower_bound.
        :param persister: A valid handle to the state store.

        :return: The next value in the range for the given lower_bound.
        """
        cur = persister.exec_stmt(
                        SHARDING_DATATYPE_HANDLER[type].SELECT_UPPER_BOUND,
                        {"raw" : False,
                        "fetch" : False,
                        "params" : (lower_bound, shard_mapping_id)})

        row = cur.fetchone()

        if row is None:
            return None

        return row[0]

    @staticmethod
    def delete_from_shard_db(table_name, type_name):
        """Delete the data from the copied data directories based on the
        sharding configuration uploaded in the sharding tables of the state
        store. The basic logic consists of

        * Querying the shard mapping ID corresponding to the sharding
          table.
        * Using the shard mapping ID to find the type of shard scheme and hence
            the sharding scheme table to query in.
        * Querying the sharding key range using the shard mapping ID.
        * Deleting the sharding keys that fall outside the range for a given
          server.

        :param table_name: The table being sharded.
        """

        shard_mapping = ShardMapping.fetch(table_name)
        if shard_mapping is None:
            raise _errors.ShardingError("Shard Mapping not found.")

        shard_mapping_id = shard_mapping.shard_mapping_id

        shards = RangeShardingSpecification.list(shard_mapping_id)
        if not shards:
            raise _errors.ShardingError("No shards associated with this"
                                                         " shard mapping ID.")

        for shard in shards:
            RangeShardingSpecification.prune_shard_id(shard.shard_id, type_name)

    @staticmethod
    def prune_shard_id(shard_id, type_name):
        """Remove the rows in the shard that do not match the metadata
        in the shard_range tables. When the rows are being removed
        foreign key checks will be disabled.

        :param shard_id: The ID of the shard that needs to be pruned.
        """

        range_sharding_spec = RangeShardingSpecification.fetch(shard_id)
        if range_sharding_spec is None:
            raise _errors.ShardingError("No shards associated with this"
                                                         " shard mapping ID.")

        shard = Shards.fetch(shard_id)

        upper_bound = RangeShardingSpecification.get_upper_bound(
                            range_sharding_spec.lower_bound,
                            range_sharding_spec.shard_mapping_id,
                            type_name
                        )

        shard_mappings = \
            ShardMapping.fetch_by_id(range_sharding_spec.shard_mapping_id)
        if shard_mappings is None:
            raise _errors.ShardingError("Shard Mapping not found.")

        #There may be multiple tables sharded by the same sharding defns. We
        #need to run prune for all of them.
        for shard_mapping in shard_mappings:
            table_name = shard_mapping.table_name

            if upper_bound is not None:
                delete_query = ("DELETE FROM %s WHERE %s < %s OR %s >= %s") % (
                    table_name,
                    shard_mapping.column_name,
                    range_sharding_spec.lower_bound,
                    shard_mapping.column_name,
                    upper_bound
                )
            else:
                delete_query = ("DELETE FROM %s WHERE %s < %s") % (
                    table_name,
                    shard_mapping.column_name,
                    range_sharding_spec.lower_bound
                )

            shard = Shards.fetch(range_sharding_spec.shard_id)
            if shard is None:
                raise _errors.ShardingError(
                    "Shard not found (%s)" %
                    (range_sharding_spec.shard_id, )
                )

            group = Group.fetch(shard.group_id)
            if group is None:
                raise _errors.ShardingError(
                    "Group not found (%s)" %
                    (shard.group_id, )
                )

            master = MySQLServer.fetch(group.master)
            if master is None:
                raise _errors.ShardingError(
                    "Group Master not found (%s)" %
                    (str(group.master))
                )

            master.connect()

            #Dependencies between tables being pruned may lead to the prune
            #failing. Hence we need to disable foreign key checks during the
            #pruning process.
            master.set_foreign_key_checks(False)
            #Perform the pruning of the table
            master.exec_stmt(delete_query)
            #Enable Foreign Key Checking
            master.set_foreign_key_checks(True)

class HashShardingSpecification(RangeShardingSpecification):
    """Represents a HASH sharding specification. The class helps encapsulate
    the representation of a typical HASH sharding implementation and is built
    upon the RANGE implementation. They share the same table in the state
    store.
    """

    #Insert a HASH of keys and the server to which they belong.
    INSERT_HASH_SPECIFICATION = (
        "INSERT INTO shard_ranges("
        "shard_mapping_id, "
        "lower_bound, "
        "shard_id) "
        "VALUES(%s, UNHEX(MD5(%s)), %s)"
    )

    #Insert Split ranges.
    #NOTE: The split lower_bound does not need the md5 algorithm.
    INSERT_HASH_SPLIT_SPECIFICATION = (
        "INSERT INTO shard_ranges("
        "shard_mapping_id, "
        "lower_bound, "
        "shard_id) "
        "VALUES(%s, UNHEX(%s), %s)"
    )

    #Given a Shard ID select the RANGE Scheme that it defines.
    SELECT_RANGE_SPECIFICATION = (
        "SELECT "
        "shard_mapping_id, "
        "HEX(lower_bound), "
        "shard_id "
        "FROM shard_ranges "
        "WHERE shard_id = %s"
    )

    #Given a Shard Mapping ID select all the RANGE mappings that it
    #defines.
    LIST_RANGE_SPECIFICATION = (
        "SELECT "
        "shard_mapping_id, "
        "HEX(lower_bound), "
        "shard_id "
        "FROM shard_ranges "
        "WHERE shard_mapping_id = %s"
    )

    #Select the least LOWER BOUND of all the LOWER BOUNDs for
    #a given shard mapping ID.
    SELECT_LEAST_LOWER_BOUND = (
        "SELECT MIN(HEX(lower_bound)) FROM "
        "shard_ranges "
        "WHERE shard_mapping_id = %s"
        )

    def __init__(self, shard_mapping_id, lower_bound, shard_id):
        """Initialize a given HASH sharding mapping specification.

        :param shard_mapping_id: The unique identification for a shard mapping.
        :param lower_bound: The lower bound of the given HASH sharding
                            definition.
        :param shard_id: An unique identification, a logical representation
                        for a shard of a particular table.
        """
        super(HashShardingSpecification, self).__init__(
            shard_mapping_id,
            lower_bound,
            shard_id
        )

    @staticmethod
    def fetch_max_key(shard_id):
        """Fetch the maximum value of the key in this shard. This will
        be used for the shards in the boundary that do not have a upper_bound
        to compare with.

        @param shard_id: The ID of the shard, from which the maximum value
                        of the tables need to be returned.

        :return: A dictionary of maximum values belonging to each of the tables
                in the shard.
        """
        #Store the list of max_keys from all the tables that belong to this
        #shard.
        max_keys = []

        hash_sharding_spec = HashShardingSpecification.fetch(shard_id)
        if hash_sharding_spec is None:
            raise _errors.ShardingError("No shards associated with this"
                                                         " shard mapping ID.")

        shard = Shards.fetch(shard_id)

        shard_mappings = \
            ShardMapping.fetch_by_id(hash_sharding_spec.shard_mapping_id)
        if shard_mappings is None:
            raise _errors.ShardingError("Shard Mapping not found.")

        for shard_mapping in shard_mappings:
            max_query = "SELECT HEX(MD5(MAX(%s))) FROM %s" % \
                        (
                        shard_mapping.column_name,
                        shard_mapping.table_name
                        )

            shard = Shards.fetch(shard_id)
            if shard is None:
                raise _errors.ShardingError(
                    "Shard not found (%s)" %
                    (hash_sharding_spec.shard_id, )
                )

            group = Group.fetch(shard.group_id)
            if group is None:
                raise _errors.ShardingError(
                    "Group not found (%s)" %
                    (shard.group_id, )
                )

            master = MySQLServer.fetch(group.master)
            if master is None:
                raise _errors.ShardingError(
                    "Group Master not found (%s)" %
                    (group.master, )
                )

            master.connect()

            cur = master.exec_stmt(
                            max_query,
                            {"raw" : False,
                            "fetch" : False,
                            }
                        )

            row = cur.fetchone()

            if row is not None:
                max_keys.append(row[0])

        #max_keys stores  all the maximum values in all the tables. We will
        #fetch the maximum values among all these values and use it as the
        #maximum values for the given shard_id.
        max_key = None
        for key in max_keys:
            if max_key is None:
                max_key = key
            elif int(key, 16) > int(max_key, 16):
                max_key = key
            else:
                pass
        #Return the maximum value among all the maximum values from
        #all the tables.
        return max_key

    @staticmethod
    def add(shard_mapping_id, shard_id, persister=None):
        """Add the HASH shard specification. This represents a single instance
        of a shard specification that maps a key HASH to a server.

        :param shard_mapping_id: The unique identification for a shard mapping.
        :param shard_id: An unique identification, a logical representation
                        for a shard of a particular table.
        """
        shard = Shards.fetch(shard_id)
        persister.exec_stmt(
            HashShardingSpecification.INSERT_HASH_SPECIFICATION, {
                "params":(
                    shard_mapping_id,
                    shard.group_id,
                    shard_id
                )
            }
        )

    @staticmethod
    def add_hash_split(shard_mapping_id, shard_id, lower_bound, persister=None):
        """Add the HASH shard specification after a split. This represents a
        single instance of a shard specification that maps a key HASH to a
        server.

        :param shard_mapping_id: The unique identification for a shard mapping.
        :param shard_id: An unique identification, a logical representation
                         for a shard of a particular table.
        :param lower_bound: The new lower_bound being inserted.
        """
        persister.exec_stmt(
            HashShardingSpecification.INSERT_HASH_SPLIT_SPECIFICATION, {
                "params":(
                    shard_mapping_id,
                    lower_bound,
                    shard_id
                )
            }
        )

    @staticmethod
    def lookup(key, shard_mapping_id, type, persister=None):
        """Return the Hash sharding specification in whose hashed key range
        the input key falls.

        :param key: The key which needs to be checked to see which range it
                    falls into
        :param shard_mapping_id: The unique identification for a shard mapping.
        :param persister: A valid handle to the state store.

        :return: The Hash Sharding Specification that contains the range in
                which the key belongs.
        """
        cur = persister.exec_stmt(SHARDING_DATATYPE_HANDLER[type].LOOKUP_KEY, {
                        "raw" : False,
                        "fetch" : False,
                        "params" : (
                            key,
                            shard_mapping_id,
                            shard_mapping_id
                        )
                    }
                )

        row = cur.fetchone()
        if row is None:
            return None
        return HashShardingSpecification(row[0], row[1],  row[2])

    @staticmethod
    def fetch(shard_id, persister=None):
        """Return the HashShardingSpecification object corresponding to the
        given sharding ID.

        :param shard_id: The unique identification for a shard.
        :param persister: A valid handle to the state store.

        :return: The HashShardingSpecification object.
                    An Empty list otherwise.
        """
        cur = persister.exec_stmt(
                    HashShardingSpecification.SELECT_RANGE_SPECIFICATION,
                        {"raw" : False,
                        "fetch" : False,
                        "params" : (shard_id,)})
        row = cur.fetchone()
        if row is None:
            return None
        return HashShardingSpecification(row[0], row[1],  row[2])


    @staticmethod
    def fetch_least_lower_bound(shard_mapping_id, persister=None):
        """Fetch the least of the lower_bound's for a given shard mapping ID.

        :param shard_mapping_id: The unique identification for a shard mapping.

        :return: The least lower_bound for a given shard mapping ID.
        """
        cur = persister.exec_stmt(
                    HashShardingSpecification.SELECT_LEAST_LOWER_BOUND,
                        {"raw" : False,
                        "fetch" : False,
                        "params" : (shard_mapping_id,)})
        row = cur.fetchone()
        if row is None:
            return None
        return row[0]

    @staticmethod
    def list(shard_mapping_id, persister=None):
        """Return the HashShardingSpecification objects corresponding to the
        given sharding scheme.

        :param shard_mapping_id: The unique identification for a shard mapping.
        :param persister: A valid handle to the state store.

        :return: A  list of HashShardingSpecification objects which belong to
                to this shard mapping.
                None if the shard mapping is not found
        """
        cur = persister.exec_stmt(
                    HashShardingSpecification.LIST_RANGE_SPECIFICATION,
                        {"raw" : False,
                        "fetch" : False,
                        "params" : (shard_mapping_id,)})
        rows = cur.fetchall()
        return [ HashShardingSpecification(*row[0:5]) for row in rows ]

    @staticmethod
    def create(persister=None):
        """We use the relations used to store the RANGE sharding data. Hence
        this method is a dummy here. We need a dummy implementation since
        the framework invokes this method to create the schemas necessary for
        storing relevant data, and we have no schemas.

        :param persister: A valid handle to the state store.
        """
        pass

    @staticmethod
    def add_constraints(persister=None):
        """Add the constraints on the sharding tables. We use a dummy
        implementation here.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        pass

    @staticmethod
    def delete_from_shard_db(table_name, type_name):
        """Delete the data from the copied data directories based on the
        sharding configuration uploaded in the sharding tables of the state
        store. The basic logic consists of

        * Querying the shard mapping ID corresponding to the sharding
          table.
        * Using the shard mapping ID to find the type of shard scheme and hence
            the sharding scheme table to query in.
        * Querying the sharding key range using the shard mapping ID.
        * Deleting the sharding keys that fall outside the range for a given
          server.

        :param table_name: The table being sharded.
        """

        shard_mapping = ShardMapping.fetch(table_name)
        if shard_mapping is None:
            raise _errors.ShardingError("Shard Mapping not found.")

        shard_mapping_id = shard_mapping.shard_mapping_id

        shards = HashShardingSpecification.list(shard_mapping_id)
        if not shards:
            raise _errors.ShardingError("No shards associated with this"
                                        " shard mapping ID.")

        for shard in shards:
            HashShardingSpecification.prune_shard_id(shard.shard_id, type_name)

    @staticmethod
    def get_upper_bound(lower_bound, shard_mapping_id, type, persister=None):
        """Return the next value in range for a given lower_bound value.
        This basically helps to form a (lower_bound, upper_bound) pair
        that can be used during a prune.

        :param lower_bound: The lower_bound value whose next range needs to
                            be retrieved.
        :param shard_mapping_id: The shard_mapping_id whose shards should be
                                searched for the given lower_bound.
        :param persister: A valid handle to the state store.

        :return: The next value in the range for the given lower_bound.
        """
        cur = persister.exec_stmt(
                        SHARDING_DATATYPE_HANDLER[type].SELECT_UPPER_BOUND,
                        {"raw" : False,
                        "fetch" : False,
                        "params" : (lower_bound, shard_mapping_id)})

        row = cur.fetchone()

        if row is None:
            return None

        return row[0]

    @staticmethod
    def prune_shard_id(shard_id, type_name):
        """Remove the rows in the shard that do not match the metadata
        in the shard_range tables.

        :param shard_id: The ID of the shard that needs to be pruned.
        """

        hash_sharding_spec = HashShardingSpecification.fetch(shard_id)
        if hash_sharding_spec is None:
            raise _errors.ShardingError("No shards associated with this"
                                                         " shard mapping ID.")

        shard = Shards.fetch(shard_id)

        upper_bound = HashShardingSpecification.get_upper_bound(
                            hash_sharding_spec.lower_bound,
                            hash_sharding_spec.shard_mapping_id,
                            type_name
                        )

        shard_mappings = \
            ShardMapping.fetch_by_id(hash_sharding_spec.shard_mapping_id)
        if shard_mappings is None:
            raise _errors.ShardingError("Shard Mapping not found.")

        for shard_mapping in shard_mappings:
            table_name = shard_mapping.table_name

            if upper_bound is not None:
                delete_query = (
                    "DELETE FROM %s "
                    "WHERE "
                    "MD5(%s) < '%s' "
                    "OR "
                    "MD5(%s) >= '%s'"
                ) % (
                    table_name,
                    shard_mapping.column_name,
                    hash_sharding_spec.lower_bound,
                    shard_mapping.column_name,
                    upper_bound
                )
            else:
                #HASH based sharding forms a circular ring. Hence
                #when there is no upper_bound, all the values, that
                #circle around from the largest lower_bound to the
                #least upper_bound need to be present in this shard.
                delete_query = ("DELETE FROM %s "
                    "WHERE "
                    "MD5(%s) >= '%s' "
                    "AND "
                    "MD5(%s) < '%s'"
                ) % (
                    table_name,
                    shard_mapping.column_name,
                    HashShardingSpecification.fetch_least_lower_bound(
                        shard_mapping.shard_mapping_id
                    ),
                    shard_mapping.column_name,
                    hash_sharding_spec.lower_bound
                )

            shard = Shards.fetch(hash_sharding_spec.shard_id)
            if shard is None:
                raise _errors.ShardingError(
                    "Shard not found (%s)" %
                    (hash_sharding_spec.shard_id, )
                )

            group = Group.fetch(shard.group_id)
            if group is None:
                raise _errors.ShardingError(
                    "Group not found (%s)" %
                    (shard.group_id, )
                )

            master = MySQLServer.fetch(group.master)
            if master is None:
                raise _errors.ShardingError(
                    "Group Master not found (%s)" %
                    (group.master, )
                )

            master.connect()

            master.set_foreign_key_checks(False)
            master.exec_stmt(delete_query)
            master.set_foreign_key_checks(True)

class MappingShardsGroups(_persistence.Persistable):
    """This class defines queries that are used to retrieve information
    on groups through a given key which could be a shard identification,
    a shard mapping identification or a table name.

    The group is used by the lock control system to avoid concurrent
    procedures simultaneously accessing the same group thus causing
    problems such as setting up a slave of a wrong master while a
    switch or fail over is going on.
    """
    # Return the local group associated with a shard_id.
    SELECT_LOCAL_GROUP_BY_SHARD_ID = \
        ("SELECT group_id FROM shards WHERE shard_id = %s")

    # Return the global group associated with a shard_id.
    SELECT_GLOBAL_GROUP_BY_SHARD_ID = \
       ("SELECT global_group AS group_id FROM shard_maps WHERE "
        "shard_mapping_id = (SELECT shard_mapping_id FROM shard_ranges WHERE "
        "shard_id = %s LIMIT 1)"
       )

    # Return the local group(s) associated with a shard_mapping_id.
    SELECT_LOCAL_GROUP_BY_SHARD_MAPPING_ID = \
       ("SELECT DISTINCT shards.group_id AS group_id FROM shard_ranges, "
        "shards WHERE shard_ranges.shard_id = shards.shard_id AND "
        "shard_ranges.shard_mapping_id = %s"
       )

    # Return the global group associated with a shard_mapping_id.
    SELECT_GLOBAL_GROUP_BY_SHARD_MAPPING_ID = \
       ("SELECT global_group AS group_id FROM shard_maps WHERE "
        "shard_mapping_id = %s"
       )

    # Return the local group(s) associated with a table_name.
    SELECT_LOCAL_GROUP_BY_TABLE_NAME = \
       ("SELECT DISTINCT shards.group_id AS group_id FROM shard_ranges, "
        "shards WHERE shard_ranges.shard_id = shards.shard_id AND "
        "shard_ranges.shard_mapping_id IN (SELECT shard_mapping_id FROM "
        "shard_tables WHERE table_name = %s)"
       )

    # Return the global group(s) associated with a table_name.
    SELECT_GLOBAL_GROUP_BY_TABLE_NAME = \
       ("SELECT global_group AS group_id FROM shard_maps WHERE "
        "shard_mapping_id = (SELECT shard_mapping_id FROM shard_tables "
        "WHERE table_name = %s)"
       )

    # Index queries through a set of pre-defined information.
    PARAM_QUERIES = {
        "local" :
        {
          "shard_id" : SELECT_LOCAL_GROUP_BY_SHARD_ID,
          "shard_mapping_id" : SELECT_LOCAL_GROUP_BY_SHARD_MAPPING_ID,
          "table_name" : SELECT_LOCAL_GROUP_BY_TABLE_NAME,
        },
        "global" :
        {
          "shard_id" : SELECT_GLOBAL_GROUP_BY_SHARD_ID,
          "shard_mapping_id" : SELECT_GLOBAL_GROUP_BY_SHARD_MAPPING_ID,
          "table_name" : SELECT_GLOBAL_GROUP_BY_TABLE_NAME,
        },
    }

    @staticmethod
    def get_group(locality, criterion, value, persister=None):
        """Return group based on a set of filters.

        :param locality: Determine if the group to be returned is global or
                         local.
        :param criterion: The criterion to retrieve a simple group: shard_id,
                          shard_mapping_id and table_name.
        :param value: Value of the criterion that will be used.
        :return: Rows with the objects requested.
        """
        assert(locality in ("local", "global"))
        assert(criterion in ("shard_id", "shard_mapping_id", "table_name"))
        rows = persister.exec_stmt(
            MappingShardsGroups.PARAM_QUERIES[locality][criterion],
            {"raw" : False, "fetch" : True, "params" : (value, )}
        )
        return rows

SHARDING_DATATYPE_HANDLER = {
    "RANGE":RangeShardingIntegerHandler,
    "HASH":HashShardingIntegerHandler
}

SHARDING_SPECIFICATION_HANDLER = {
    "RANGE":RangeShardingSpecification,
    "HASH":HashShardingSpecification
}



#!/usr/bin/env python
#
# Copyright (c) 2011, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This file contains the object diff utility which allows users to compare the
definitions of two objects and return the difference (like diff).
"""

from mysql.utilities.common.tools import check_python_version

# Check Python version compatibility
check_python_version()

import os
import re
import sys

from mysql.utilities.exception import FormatError, UtilError
from mysql.utilities.command.diff import object_diff, database_diff
from mysql.utilities.common.ip_parser import parse_connection
from mysql.utilities.common.tools import check_connector_python
from mysql.utilities.common.pattern_matching import REGEXP_QUALIFIED_OBJ_NAME
from mysql.utilities.common.messages import (PARSE_ERR_DB_OBJ_MISSING,
                                             PARSE_ERR_DB_OBJ_MISSING_MSG,
                                             PARSE_ERR_DB_OBJ_PAIR,
                                             PARSE_ERR_DB_OBJ_PAIR_EXT)
from mysql.utilities.common.options import (add_difftype, add_verbosity,
                                            check_verbosity, add_changes_for,
                                            add_reverse, setup_common_options,
                                            add_character_set_option)


# Constants
NAME = "MySQL Utilities - mysqldiff "
DESCRIPTION = "mysqldiff - compare object definitions among objects" + \
              " where the difference is how db1.obj1 differs from db2.obj2"
USAGE = "%prog --server1=user:pass@host:port:socket " + \
        "--server2=user:pass@host:port:socket db1.object1:db2.object1 db3:db4"
PRINT_WIDTH = 75

# Check for connector/python
if not check_connector_python():
    sys.exit(1)

if __name__ == '__main__':
    # Setup the command parser
    parser = setup_common_options(os.path.basename(sys.argv[0]),
                                  DESCRIPTION, USAGE, False, False)

    # Connection information for the source server
    parser.add_option("--server1", action="store", dest="server1",
                      type="string", default="root@localhost:3306",
                      help="connection information for first server in the "
                           "form: <user>[:<password>]@<host>[:<port>]"
                           "[:<socket>] or <login-path>[:<port>][:<socket>].")

    # Connection information for the destination server
    parser.add_option("--server2", action="store", dest="server2",
                      type="string", default=None,
                      help="connection information for second server in the "
                           "form: <user>[:<password>]@<host>[:<port>]"
                           "[:<socket>] or <login-path>[:<port>][:<socket>].")

    # Add character set option
    add_character_set_option(parser)

    # Add display width option
    parser.add_option("--width", action="store", dest="width",
                      type="int", help="display width",
                      default=PRINT_WIDTH)

    # Force mode
    parser.add_option("--force", action="store_true", dest="force",
                      help="do not abort when a diff test fails")

    # Add compact option for resulting diff
    parser.add_option("-c", "--compact", action="store_true",
                      dest="compact", help="compact output from a diff.")

    # Skip check of table options.
    parser.add_option("--skip-table-options", action="store_true",
                      dest="skip_tbl_opts",
                      help="skip check of all table options (e.g., "
                           "AUTO_INCREMENT, ENGINE, CHARSET, etc.).")

    # Add verbosity and quiet (silent) mode
    add_verbosity(parser, True)

    # Add difftype option with SQL option
    add_difftype(parser, True)

    # Add the direction (changes-for)
    add_changes_for(parser, None)

    # Add show reverse option
    add_reverse(parser)

    # Now we process the rest of the arguments.
    opt, args = parser.parse_args()

    # Warn if quiet and verbosity are both specified
    check_verbosity(opt)

    # Set options for database operations.
    options = {
        "quiet": opt.quiet,
        "verbosity": opt.verbosity,
        "difftype": opt.difftype,
        "force": opt.force,
        "width": opt.width,
        "changes-for": opt.changes_for,
        "reverse": opt.reverse,
        "skip_table_opts": opt.skip_tbl_opts,
        "compact": opt.compact,
        "charset": opt.charset,
    }

    # Parse server connection values
    try:
        server1_values = parse_connection(opt.server1, None, options)
    except FormatError:
        _, err, _ = sys.exc_info()
        parser.error("Server1 connection values invalid: %s." % err)
    except UtilError:
        _, err, _ = sys.exc_info()
        parser.error("Server1 connection values invalid: %s." % err.errmsg)
    if opt.server2 is not None:
        try:
            server2_values = parse_connection(opt.server2, None, options)
        except FormatError:
            _, err, _ = sys.exc_info()
            parser.error("Server2 connection values invalid: %s." % err)
        except UtilError:
            _, err, _ = sys.exc_info()
            parser.error("Server2 connection values invalid: %s." % err.errmsg)
    else:
        server2_values = None

    # Check for arguments
    if len(args) == 0:
        parser.error("No objects specified to compare.")

    # run the diff
    diff_failed = False
    arg_regexp = re.compile(r'{0}(?:\:){0}'.format(REGEXP_QUALIFIED_OBJ_NAME))
    for argument in args:
        m_obj = arg_regexp.match(argument)
        if not m_obj:
            parser.error(PARSE_ERR_DB_OBJ_PAIR.format(db_obj_pair=argument,
                                                      db1_label='db1',
                                                      obj1_label='object1',
                                                      db2_label='db2',
                                                      obj2_label='object2'))
        db1, obj1, db2, obj2 = m_obj.groups()

        # Verify if the size of the objects matched by the REGEX is equal to
        # the initial specified string. In general, this identifies the
        # missing use of backticks.
        matched_size = len(db1)
        if obj1:
            # add 1 for the separator '.'
            matched_size = matched_size + 1
            matched_size = matched_size + len(obj1)
        # add 1 for the separator ':'
        matched_size = matched_size + 1
        matched_size = matched_size + len(db2)
        if obj2:
            # add 1 for the separator '.'
            matched_size = matched_size + 1
            matched_size = matched_size + len(obj2)
        if matched_size != len(argument):
            parser.error(PARSE_ERR_DB_OBJ_PAIR_EXT.format(db_obj_pair=argument,
                                                          db1_label='db1',
                                                          obj1_label='object1',
                                                          db2_label='db2',
                                                          obj2_label='object2',
                                                          db1_value=db1,
                                                          obj1_value=obj1,
                                                          db2_value=db2,
                                                          obj2_value=obj2))

        if (obj1 and not obj2) or (not obj1 and obj2):
            if obj1:
                detail = PARSE_ERR_DB_OBJ_MISSING.format(db_no_obj_label='db2',
                                                         db_no_obj_value=db2,
                                                         only_obj_value=obj1,
                                                         db_obj_label='db1',
                                                         db_obj_value=db1)
            else:
                detail = PARSE_ERR_DB_OBJ_MISSING.format(db_no_obj_label='db1',
                                                         db_no_obj_value=db1,
                                                         only_obj_value=obj2,
                                                         db_obj_label='db2',
                                                         db_obj_value=db2)
            parser.error(PARSE_ERR_DB_OBJ_MISSING_MSG.format(
                detail=detail,
                db1_label='db1',
                obj1_label='object1',
                db2_label='db2',
                obj2_label='object2'))

        # We have db1.obj:db2.obj
        if obj1:
            try:
                diff = object_diff(server1_values, server2_values,
                                   "%s.%s" % (db1, obj1),
                                   "%s.%s" % (db2, obj2), options)
            except UtilError:
                _, e, _ = sys.exc_info()
                print("ERROR: %s" % e.errmsg)
                sys.exit(1)
            if diff is not None:
                diff_failed = True

        # We have db1:db2
        else:
            try:
                res = database_diff(server1_values, server2_values,
                                    db1, db2, options)
            except UtilError:
                _, e, _ = sys.exc_info()
                print("ERROR: %s" % e.errmsg)
                sys.exit(1)
            if not res:
                diff_failed = True

    if diff_failed:
        if not opt.quiet:
            print("Compare failed. One or more differences found.")
        sys.exit(1)

    if not opt.quiet:
        print("Success. All objects are the same.")

    sys.exit()



#
# Copyright (c) 2010, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This module contains the following methods design to support common operations
over the ip address or hostnames among the multiple utilities.

Methods:
  parse_connection()         Parse connection parameters
"""

import re
import os
import logging

from mysql.utilities.exception import UtilError, FormatError
from mysql.utilities.common.my_print_defaults import (MyDefaultsReader,
                                                      my_login_config_exists,
                                                      my_login_config_path)


log = logging.getLogger('ip_parser')

_BAD_CONN_FORMAT = (u"Connection '{0}' cannot be parsed. Please review the "
                    u"used connection string (accepted formats: "
                    u"<user>[:<password>]@<host>[:<port>][:<socket>] or "
                    u"<login-path>[:<port>][:<socket>])")

_BAD_QUOTED_HOST = u"Connection '{0}' has a malformed quoted host"

_UNPARSED_CONN_FORMAT = ("Connection '{0}' not parsed completely. Parsed "
                         "elements '{1}', unparsed elements '{2}'")

_CONN_USERPASS = re.compile(
    r"(?P<fquote>[\'\"]?)"    # First quote
    r"(?P<user>.+?)"          # User name
    r"(?:(?P=fquote))"        # First quote match
    r"(?:\:"                  # Optional :
    r"(?P<squote>[\'\"]?)"    # Second quote
    r"(?P<passwd>.+)"         # Password
    r"(?P=squote))"           # Second quote match
    r"|(?P<sfquote>[\'\"]?)"  # Quote on single user name
    r"(?P<suser>.+)"          # Single user name
    r"(?:(?P=sfquote))"       # Quote match on single user name
)

_CONN_QUOTEDHOST = re.compile(
    r"((?:^[\'].*[\'])|(?:^[\"].*[\"]))"  # quoted host name
    r"(?:\:(\d+))?"                       # Optional port number
    r"(?:\:([\/\\w+.\w+.\-]+))?"          # Optional path to socket
)

_CONN_LOGINPATH = re.compile(
    r"((?:\\\"|[^:])+|(?:\\\'|[^:])+)"  # login-path
    r"(?:\:(\d+))?"                     # Optional port number
    r"(?:\:([\/\\w+.\w+.\-]+))?"        # Optional path to socket
)

_CONN_ANY_HOST = re.compile(
    r"""([\w\.]*%)
       (?:\:{0,1}(.*))                   # capture all the rest
    """, re.VERBOSE)

_CONN_HOST_NAME = re.compile(
    r"""(
        (?:
           (?:
              (?:
                 (?!-)         # must not start with hyphen '-'
                 (?:[\w\d-])*  # must not end with the hyphen
                 [A-Za-z]      # starts with a character from the alphabet
                 (?:[\w\d-])*
                 (?:
                    (?<!-)     # end capturing if a '-' is followed by '.'
                 )
               ){1,63}         # limited length for segment
            )
         (?:                   # following segments
            (?:\.)?            # the segment separator  the dot '.'
            (?:
               (?!-)
               [\w\d-]{1,63}   # last segment
               (?<!-)          #shuld not end with hyphen
            )
          )*
         )
        )
       (.*)                    # capture all the rest
     """, re.VERBOSE)

_CONN_IPv4_NUM_ONLY = re.compile(
    r"""(
          (?:         # start of the IPv4 1st group
             25[0-4]  # this match numbers 250 to 254
                    | # or
             2[0-4]\d # this match numbers from 200 to 249
                    | # or
             1\d\d    # this match numbers from 100 to 199
                    | # or
             [1-9]{0,1}\d # this match numbers from 0 to 99
           )
          (?:         # start of the 3 next groups
             \.       # the prefix '.' like in '.255'
             (?:
                25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d
                      # same group as before
              )
           )
             {3}      # but it will match 3 times of it and prefixed by '.'
          )
          (?:\:{0,1}(.*))
          """, re.VERBOSE)

_CONN_port_ONLY = re.compile(
    r"""(?:
          \]{0,1}             # the ']' of IPv6 -optional
                 \:{0,1}      # the ':' for port number  -optional
                        (
                         \d*  # matches any sequence of numbers
                         )
         )          # end of port number group
        (?:\:{0,1}(.*))      # all the rest to extract the socket
        """, re.VERBOSE)

_CONN_socket_ONLY = re.compile(
    r"""(?:           # Not capturing group of ':'
           \:{0,1}
             ([      # Capturing '\' or '/' file name.ext
               \/\\w+.\w+.\-
               ]+    # to match a path
              )
        )?
       (.*)          # all the rest to advice the user.
    """, re.VERBOSE)

_CONN_IPv6 = re.compile(
    r"""
    \[{0,1}                   # the optional heading '['
    (
     (?!.*::.*::)              # Only a single whildcard allowed
     (?:(?!:)|:(?=:))          # Colon iff it would be part of a wildcard
     (?:                       # Repeat 6 times:
        [0-9a-f]{0,4}          # A group of at most four hexadecimal digits
        (?:(?<=::)|(?<!::):)   # Colon unless preceded by wildcard
     ){6}                      # expecting 6 groups
     (?:                       # Either
        [0-9a-f]{0,4}          # Another group
        (?:(?<=::)|(?<!::):)   # Colon unless preceded by wildcard
        [0-9a-f]{0,4}          # Last group
        (?:(?<=::)             # Colon iff preceded by exacly one colon
           |(?<!:)
           |(?<=:)(?<!::):
         )
      )
     )
     (?:
        \]{0,1}\:{0,1}(.*)     # optional closing ']' and group for the rest
      )
    """, re.VERBOSE)

# Type of address amd Key names for the dictionary IP_matchers
HN = "hostname"
ipv4 = "IPv4"
ipv6 = "IPv6"
ANY_LIKE = "host like"
# This list is used to set an order to the matchers.
IP_matchers_list = [ipv4, ipv6, ANY_LIKE, HN]
# This dictionary is used to identify the matched type..
IP_matchers = {
    ANY_LIKE: _CONN_ANY_HOST,
    HN: _CONN_HOST_NAME,
    ipv4: _CONN_IPv4_NUM_ONLY,
    ipv6: _CONN_IPv6
}


def hostname_is_ip(hostname):
    """Determine hostname is an IP address.

    Return bool - True = is IP address
    """
    if len(hostname.split(":")) <= 1:  # if fewer colons, must be IPv4
        grp = _CONN_IPv4_NUM_ONLY.match(hostname)
    else:
        grp = _CONN_IPv6.match(hostname)
    if not grp:
        return False
    return True


def parse_connection(connection_values, my_defaults_reader=None, options=None):
    """Parse connection values.

    The function parses a connection specification of one of the forms::

      - user[:password]@host[:port][:socket]
      - login-path[:port][:socket]

    A dictionary is returned containing the connection parameters. The
    function is designed so that it shall be possible to use it with a
    ``connect`` call in the following manner::

      options = parse_connection(spec)
      conn = mysql.connector.connect(**options)

    conn_values[in]         Connection values in the form:
                            user:password@host:port:socket
                            or login-path:port:socket
    my_defaults_reader[in]  Instance of MyDefaultsReader to read the
                            information of the login-path from configuration
                            files. By default, the value is None.
    options[in]             Dictionary of options (e.g. basedir), from the used
                            utility. By default, it set with an empty
                            dictionary. Note: also supports options values
                            from optparse.

    Notes:

    This method validates IPv4 addresses and standard IPv6 addresses.

    This method accepts quoted host portion strings. If the host is marked
    with quotes, the code extracts this without validation and assigns it to
    the host variable in the returned tuple. This allows users to specify host
    names and IP addresses that are outside of the supported validation.

    Returns dictionary (user, passwd, host, port, socket)
            or raise an exception if parsing error
    """
    if options is None:
        options = {}

    def _match(pattern, search_str):
        """Returns the groups from string search or raise FormatError if it
        does not match with the pattern.
        """
        grp = pattern.match(search_str)
        if not grp:
            raise FormatError(_BAD_CONN_FORMAT.format(connection_values))
        return grp.groups()

    # Split on the '@' to determine the connection string format.
    # The user/password may have the '@' character, split by last occurrence.
    conn_format = connection_values.rsplit('@', 1)

    if len(conn_format) == 1:
        # No '@' then handle has in the format: login-path[:port][:socket]
        login_path, port, socket = _match(_CONN_LOGINPATH, conn_format[0])

        #Check if the login configuration file (.mylogin.cnf) exists
        if login_path and not my_login_config_exists():
            raise UtilError(".mylogin.cnf was not found at is default "
                            "location: %s."
                            "Please configure your login-path data before "
                            "using it (use the mysql_config_editor tool)."
                            % my_login_config_path())

        # If needed, create a MyDefaultsReader and search for my_print_defaults
        # tool.
        if not my_defaults_reader:
            my_defaults_reader = MyDefaultsReader(options)
        elif not my_defaults_reader.tool_path:
            my_defaults_reader.search_my_print_defaults_tool()

        # Check if the my_print_default tool is able to read a login-path from
        # the mylogin configuration file
        if not my_defaults_reader.check_login_path_support():
            raise UtilError("the used my_print_defaults tool does not "
                            "support login-path options: %s. "
                            "Please confirm that the location to a tool with "
                            "login-path support is included in the PATH "
                            "(at the beginning)."
                            % my_defaults_reader.tool_path)

        # Read and parse the login-path data (i.e., user, password and host)
        login_path_data = my_defaults_reader.get_group_data(login_path)

        if login_path_data:
            user = login_path_data.get('user', None)
            passwd = login_path_data.get('password', None)
            host = login_path_data.get('host', None)
            if not port:
                port = login_path_data.get('port', None)
            if not socket:
                socket = login_path_data.get('socket', None)

            if os.name == "posix" and socket is not None:
                # if we are on unix systems and used a socket, hostname can be
                # safely assumed as being localhost so it is not required
                required_options = ('user', 'socket')
                host = 'localhost' if host is None else host
            else:
                required_options = ('user', 'host', 'port')

            missing_options = [opt for opt in required_options
                               if locals()[opt] is None]
            # If we are on unix and port is missing, user might have specified
            # a socket instead
            if os.name == "posix" and "port" in missing_options:
                i = missing_options.index("port")
                if socket:  # If we have a socket, port is not needed
                    missing_options.pop(i)
                else:
                    # if we don't have neither a port nor a socket, we need
                    # either a port or a socket
                    missing_options[i] = "port or socket"

            if missing_options:
                message = ",".join(missing_options)
                if len(missing_options) > 1:
                    comma_idx = message.rfind(",")
                    message = "{0} and {1}".format(message[:comma_idx],
                                                   message[comma_idx + 1:])
                pluralize = "s" if len(missing_options) > 1 else ""
                raise UtilError("Missing connection value{0} for "
                                "{1} option{0}".format(pluralize, message))
        else:
            raise UtilError("No login credentials found for login-path: %s. "
                            "Please review the used connection string: %s"
                            % (login_path, connection_values))

    elif len(conn_format) == 2:

        # Handle as in the format: user[:password]@host[:port][:socket]
        userpass, hostportsock = conn_format

        # Get user, password
        match = _CONN_USERPASS.match(userpass)
        if not match:
            raise FormatError(_BAD_CONN_FORMAT.format(connection_values))
        user = match.group('user')
        if user is None:
            # No password provided
            user = match.group('suser').rstrip(':')
        passwd = match.group('passwd')

        # Handle host, port and socket
        if len(hostportsock) <= 0:
            raise FormatError(_BAD_CONN_FORMAT.format(connection_values))

        if hostportsock[0] in ['"', "'"]:
            # need to strip the quotes
            host, port, socket = _match(_CONN_QUOTEDHOST, hostportsock)
            if host[0] == '"':
                host = host.strip('"')
            if host[0] == "'":
                host = host.strip("'")

        else:
            host, port, socket, _ = parse_server_address(hostportsock)

    else:
        # Unrecognized format
        raise FormatError(_BAD_CONN_FORMAT.format(connection_values))

    # Get character-set from options
    if isinstance(options, dict):
        charset = options.get("charset", None)
    else:
        # options is an instance of optparse.Values
        try:
            charset = options.charset  # pylint: disable=E1103
        except AttributeError:
            charset = None

    # Set parsed connection values
    connection = {
        "user": user,
        "host": host,
        "port": int(port) if port else 3306,
        "passwd": passwd if passwd else ''
    }

    if charset:
        connection["charset"] = charset

    # Handle optional parameters. They are only stored in the dict if
    # they were provided in the specifier.
    if socket is not None and os.name == "posix":
        connection['unix_socket'] = socket

    return connection


def parse_server_address(connection_str):
    """Parses host, port and socket from the given connection string.

    Returns a tuple of (host, port, socket, add_type) where add_type is
    the name of the parser that successfully parsed the hostname from
    the connection string.
    """
    # Default values to return.
    host = None
    port = None
    socket = None
    address_type = None
    unparsed = None
    # From the matchers look the one that match a host.
    for IP_matcher in IP_matchers_list:
        try:
            group = _match(IP_matchers[IP_matcher], connection_str)
            if group:
                host = group[0]
                if IP_matcher == ipv6:
                    host = "[%s]" % host

                if group[1]:
                    part2_port_socket = _match(_CONN_port_ONLY, group[1],
                                               trow_error=False)
                    if not part2_port_socket:
                        unparsed = group[1]
                    else:
                        port = part2_port_socket[0]
                        if part2_port_socket[1]:
                            part4 = _match(_CONN_socket_ONLY,
                                           part2_port_socket[1],
                                           trow_error=False)
                            if not part4:
                                unparsed = part2_port_socket[1]
                            else:
                                socket = part4[0]
                                unparsed = part4[1]

            # If host is match we stop looking as is the most significant.
            if host:
                address_type = IP_matcher
                break
        # ignore the error trying to match.
        except FormatError:
            pass
    # we must alert, that the connection could not be parsed.
    if host is None:
        raise FormatError(_BAD_CONN_FORMAT.format(connection_str))
    _verify_parsing(connection_str, host, port, socket, address_type, unparsed)

    return host, port, socket, address_type


def _verify_parsing(connection_str, host, port, socket, address_type,
                    unparsed):
    """Verify that the connection string was totally parsed and not parts of
    it where not matched, otherwise raise an error.
    """
    exp_connection_str = connection_str
    log.debug("exp_connection_str {0}".format(exp_connection_str))
    parsed_connection_list = []
    if host:
        log.debug("host {0}".format(host))
        if address_type == ipv6 and not "[" in connection_str:
            host = host.replace("[", "")
            host = host.replace("]", "")
        parsed_connection_list.append(host)
    if port:
        log.debug("port {0}".format(port))
        parsed_connection_list.append(port)
    if socket:
        log.debug("socket {0}".format(socket))
        parsed_connection_list.append(socket)
    parsed_connection = ":".join(parsed_connection_list)
    log.debug('parsed_connection {0}'.format(parsed_connection))
    diff = None
    if not unparsed:
        log.debug('not unparsed found, creating diff')
        diff = connection_str.replace(host, "")
        if port:
            diff = diff.replace(port, "")
        if socket:
            diff = diff.replace(socket, "")
        log.debug("diff {0}".format(diff))
    log.debug("unparsed {0}".format(unparsed))
    if unparsed or (exp_connection_str != parsed_connection
                    and (diff and diff != ":")):
        log.debug("raising exception")
        parsed_args = "host:%s, port:%s, socket:%s" % (host, port, socket)
        log.debug(_UNPARSED_CONN_FORMAT.format(connection_str,
                                               parsed_args,
                                               unparsed))
        raise FormatError(_UNPARSED_CONN_FORMAT.format(connection_str,
                                                       parsed_args,
                                                       unparsed))


def _match(pattern, connection_str, trow_error=True):
    """Tries to match a pattern with the connection string and returns the
    groups.
    """
    grp = pattern.match(connection_str)
    if not grp:
        if trow_error:
            raise FormatError(_BAD_CONN_FORMAT.format(connection_str))
        return False
    return grp.groups()


def clean_IPv6(host_address):
    """Clean IPv6 host address
    """
    if host_address:
        host_address = host_address.replace("[", "")
        host_address = host_address.replace("]", "")
    return host_address


def format_IPv6(host_address):
    """Format IPv6 host address
    """
    if host_address:
        if not "]" in host_address:
            host_address = "[{0}]".format(host_address)
    return host_address



#
# Copyright (c) 2010, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This file contains the copy database operation which ensures a database
is exactly the same among two servers.
"""

import sys

from mysql.utilities.exception import UtilError
from mysql.utilities.common.database import Database
from mysql.utilities.common.options import check_engine_options
from mysql.utilities.common.server import connect_servers
from mysql.utilities.command.dbexport import (get_change_master_command,
                                              get_copy_lock, get_gtid_commands)


_RPL_COMMANDS, _RPL_FILE = 0, 1

_GTID_WARNING = ("# WARNING: The server supports GTIDs but you have elected "
                 "to skip exexcuting the GTID_EXECUTED statement. Please "
                 "refer to the MySQL online reference manual for more "
                 "information about how to handle GTID enabled servers with "
                 "backup and restore operations.")
_GTID_BACKUP_WARNING = ("# WARNING: A partial copy from a server that has "
                        "GTIDs enabled will by default include the GTIDs of "
                        "all transactions, even those that changed suppressed "
                        "parts of the database. If you don't want to generate "
                        "the GTID statement, use the --skip-gtid option. To "
                        "export all databases, use the --all option and do "
                        "not specify a list of databases.")
_NON_GTID_WARNING = ("# WARNING: The %s server does not support GTIDs yet the "
                     "%s server does support GTIDs. To suppress this warning, "
                     "use the --skip-gtid option when copying %s a non-GTID "
                     "enabled server.")


def _copy_objects(source, destination, db_list, options,
                  show_message=True, do_create=True):
    """Copy objects for a list of databases

    This method loops through a list of databases copying the objects as
    controlled by the skip options.

    source[in]             Server class instance for source
    destination[in]        Server class instance for destination
    options[in]            copy options
    show_message[in]       if True, display copy message
                           Default = True
    do_create[in]          if True, execute create statement for database
                           Default = True
    """
    # Copy objects
    for db_name in db_list:

        if show_message:
            # Display copy message
            if not options.get('quiet', False):
                msg = "# Copying database %s " % db_name[0]
                if db_name[1]:
                    msg += "renamed as %s" % (db_name[1])
                print msg

        # Get a Database class instance
        db = Database(source, db_name[0], options)

        # Perform the copy
        db.init()
        db.copy_objects(db_name[1], options, destination,
                        options.get("threads", False), do_create)


def multiprocess_db_copy_task(copy_db_task):
    """Multiprocess copy database method.

    This method wraps the copy_db method to allow its concurrent
    execution by a pool of processes.

    copy_db_task[in]    dictionary of values required by a process to perform
                        the database copy task, namely:
                        {'source_srv': <dict with source connect values>,
                         'dest_srv': <dict with destination connect values>,
                         'db_list': <list of databases to copy>,
                         'options': <dict of options>,
                        }
    """
    # Get input values to execute task.
    source_srv = copy_db_task.get('source_srv')
    dest_srv = copy_db_task.get('dest_srv')
    db_list = copy_db_task.get('db_list')
    options = copy_db_task.get('options')
    # Execute copy databases task.
    # NOTE: Must handle any exception here, because worker processes will not
    # propagate them to the main process.
    try:
        copy_db(source_srv, dest_srv, db_list, options)
    except UtilError:
        _, err, _ = sys.exc_info()
        print("ERROR: {0}".format(err.errmsg))


def copy_db(src_val, dest_val, db_list, options):
    """Copy a database

    This method will copy a database and all of its objects and data from
    one server (source) to another (destination). Options are available to
    selectively ignore each type of object. The do_drop parameter is
    used to permit the copy to overwrite an existing destination database
    (default is to not overwrite).

    src_val[in]        a dictionary containing connection information for the
                       source including:
                       (user, password, host, port, socket)
    dest_val[in]       a dictionary containing connection information for the
                       destination including:
                       (user, password, host, port, socket)
    options[in]        a dictionary containing the options for the copy:
                       (skip_tables, skip_views, skip_triggers, skip_procs,
                       skip_funcs, skip_events, skip_grants, skip_create,
                       skip_data, verbose, do_drop, quiet,
                       connections, debug, exclude_names, exclude_patterns)

    Notes:
        do_drop  - if True, the database on the destination will be dropped
                   if it exists (default is False)
        quiet    - do not print any information during operation
                   (default is False)

    Returns bool True = success, False = error
    """
    verbose = options.get("verbose", False)
    quiet = options.get("quiet", False)
    do_drop = options.get("do_drop", False)
    skip_views = options.get("skip_views", False)
    skip_procs = options.get("skip_procs", False)
    skip_funcs = options.get("skip_funcs", False)
    skip_events = options.get("skip_events", False)
    skip_grants = options.get("skip_grants", False)
    skip_data = options.get("skip_data", False)
    skip_triggers = options.get("skip_triggers", False)
    skip_tables = options.get("skip_tables", False)
    skip_gtid = options.get("skip_gtid", False)
    locking = options.get("locking", "snapshot")

    conn_options = {
        'quiet': quiet,
        'version': "5.1.30",
    }
    servers = connect_servers(src_val, dest_val, conn_options)
    cloning = (src_val == dest_val) or dest_val is None

    source = servers[0]
    if cloning:
        destination = servers[0]
    else:
        destination = servers[1]

    src_gtid = source.supports_gtid() == 'ON'
    dest_gtid = destination.supports_gtid() == 'ON'if destination else False

    # Get list of all databases from source if --all is specified.
    # Ignore system databases.
    if options.get("all", False):
        # The --all option is valid only if not cloning.
        if not cloning:
            if not quiet:
                print "# Including all databases."
            rows = source.get_all_databases()
            for row in rows:
                db_list.append((row[0], None))  # Keep same name
        else:
            raise UtilError("Cannot copy all databases on the same server.")
    elif not skip_gtid and src_gtid:
        # Check to see if this is a full copy (complete backup)
        all_dbs = source.exec_query("SHOW DATABASES")
        dbs = [db[0] for db in db_list]
        for db in all_dbs:
            if db[0].upper() in ["MYSQL", "INFORMATION_SCHEMA",
                                 "PERFORMANCE_SCHEMA"]:
                continue
            if not db[0] in dbs:
                print _GTID_BACKUP_WARNING
                break

    # Do error checking and preliminary work:
    #  - Check user permissions on source and destination for all databases
    #  - Check to see if executing on same server but same db name (error)
    #  - Build list of tables to lock for copying data (if no skipping data)
    #  - Check storage engine compatibility
    for db_name in db_list:
        source_db = Database(source, db_name[0])
        if destination is None:
            destination = source
        if db_name[1] is None:
            db = db_name[0]
        else:
            db = db_name[1]
        dest_db = Database(destination, db)

        # Make a dictionary of the options
        access_options = {
            'skip_views': skip_views,
            'skip_procs': skip_procs,
            'skip_funcs': skip_funcs,
            'skip_grants': skip_grants,
            'skip_events': skip_events,
            'skip_triggers': skip_triggers,
        }

        source_db.check_read_access(src_val["user"], src_val["host"],
                                    access_options)

        # Make a dictionary containing the list of objects from source db
        source_objects = {
            "views": source_db.get_db_objects("VIEW", columns="full"),
            "procs": source_db.get_db_objects("PROCEDURE", columns="full"),
            "funcs": source_db.get_db_objects("FUNCTION", columns="full"),
            "events": source_db.get_db_objects("EVENT", columns="full"),
            "triggers": source_db.get_db_objects("TRIGGER", columns="full"),
        }

        dest_db.check_write_access(dest_val['user'], dest_val['host'],
                                   access_options, source_objects, do_drop)

        # Error is source db and destination db are the same and we're cloning
        if destination == source and db_name[0] == db_name[1]:
            raise UtilError("Destination database name is same as "
                            "source - source = %s, destination = %s" %
                            (db_name[0], db_name[1]))

        # Error is source database does not exist
        if not source_db.exists():
            raise UtilError("Source database does not exist - %s" % db_name[0])

        # Check storage engines
        check_engine_options(destination,
                             options.get("new_engine", None),
                             options.get("def_engine", None),
                             False, options.get("quiet", False))

    # Get replication commands if rpl_mode specified.
    # if --rpl specified, dump replication initial commands
    rpl_info = None

    # Turn off foreign keys if they were on at the start
    destination.disable_foreign_key_checks(True)

    # Get GTID commands
    if not skip_gtid:
        gtid_info = get_gtid_commands(source)
        if src_gtid and not dest_gtid:
            print _NON_GTID_WARNING % ("destination", "source", "to")
        elif not src_gtid and dest_gtid:
            print _NON_GTID_WARNING % ("source", "destination", "from")
    else:
        gtid_info = None
        if src_gtid and not cloning:
            print _GTID_WARNING

    # If cloning, turn off gtid generation
    if gtid_info and cloning:
        gtid_info = None
    # if GTIDs enabled, write the GTID commands
    if gtid_info and dest_gtid:
        # Check GTID version for complete feature support
        destination.check_gtid_version()
        # Check the gtid_purged value too
        destination.check_gtid_executed()
        for cmd in gtid_info[0]:
            print "# GTID operation:", cmd
            destination.exec_query(cmd, {'fetch': False, 'commit': False})

    if options.get("rpl_mode", None):
        new_opts = options.copy()
        new_opts['multiline'] = False
        new_opts['strict'] = True
        rpl_info = get_change_master_command(src_val, new_opts)
        destination.exec_query("STOP SLAVE", {'fetch': False, 'commit': False})

    # Copy (create) objects.
    # We need to delay trigger and events to after data is loaded
    new_opts = options.copy()
    new_opts['skip_triggers'] = True
    new_opts['skip_events'] = True

    # Get the table locks unless we are cloning with lock-all
    if not (cloning and locking == 'lock-all'):
        my_lock = get_copy_lock(source, db_list, options, True)

    _copy_objects(source, destination, db_list, new_opts)

    # If we are cloning, take the write locks prior to copying data
    if cloning and locking == 'lock-all':
        my_lock = get_copy_lock(source, db_list, options, True, cloning)

    # Copy tables data
    if not skip_data and not skip_tables:

        # Copy tables
        for db_name in db_list:

            # Get a Database class instance
            db = Database(source, db_name[0], options)

            # Perform the copy
            # Note: No longer use threads, use multiprocessing instead.
            db.init()
            db.copy_data(db_name[1], options, destination, connections=1,
                         src_con_val=src_val, dest_con_val=dest_val)

    # if cloning with lock-all unlock here to avoid system table lock conflicts
    if cloning and locking == 'lock-all':
        my_lock.unlock()

    # Create triggers for all databases
    if not skip_triggers:
        new_opts = options.copy()
        new_opts['skip_tables'] = True
        new_opts['skip_views'] = True
        new_opts['skip_procs'] = True
        new_opts['skip_funcs'] = True
        new_opts['skip_events'] = True
        new_opts['skip_grants'] = True
        new_opts['skip_create'] = True
        _copy_objects(source, destination, db_list, new_opts, False, False)

    # Create events for all databases
    if not skip_events:
        new_opts = options.copy()
        new_opts['skip_tables'] = True
        new_opts['skip_views'] = True
        new_opts['skip_procs'] = True
        new_opts['skip_funcs'] = True
        new_opts['skip_triggers'] = True
        new_opts['skip_grants'] = True
        new_opts['skip_create'] = True
        _copy_objects(source, destination, db_list, new_opts, False, False)

    if not (cloning and locking == 'lock-all'):
        my_lock.unlock()

    # if GTIDs enabled, write the GTID-related commands
    if gtid_info and dest_gtid:
        print "# GTID operation:", gtid_info[1]
        destination.exec_query(gtid_info[1])

    if options.get("rpl_mode", None):
        for cmd in rpl_info[_RPL_COMMANDS]:
            if cmd[0] == '#' and not quiet:
                print cmd
            else:
                if verbose:
                    print cmd
                destination.exec_query(cmd)
        destination.exec_query("START SLAVE;")

    # Turn on foreign keys if they were on at the start
    destination.disable_foreign_key_checks(False)

    if not quiet:
        print "#...done."
    return True



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Define features that can be used throughout the code.
"""
import os
import sys
import inspect
import ctypes
import re
import datetime

TTL = 0
VERSION_TOKEN = 0
FABRIC_UUID = 0

class SingletonMeta(type):
    """Define a Singleton.
    This Singleton class can be used as follows::

      class MyClass(object):
        __metaclass__ = SingletonMeta
      ...
    """
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonMeta, cls).__call__(*args,
                                                                     **kwargs)
        return cls._instances[cls]


class Singleton(object):
    """Define a Singleton.
    This Singleton class can be used as follows::

      class MyClass(Singleton):
      ...
    """
    __metaclass__ = SingletonMeta

def _do_fork():
    """Create a process.
    """
    try:
        if os.fork() > 0:
            sys.exit(0)
    except OSError, error:
        sys.stderr.write("fork failed with errno %d: %s\n" %
                         (error.errno, error.strerror))
        sys.exit(1)

def daemonize(stdin='/dev/null', stdout='/dev/null', stderr='/dev/null'):
    """Standard procedure for daemonizing a process.

    This process daemonizes the current process and put it in the
    background. When daemonized, logs are written to syslog.

    [1] Python Cookbook by Martelli, Ravenscropt, and Ascher.
    """
    _do_fork()
    os.chdir("/")        # The current directory might be removed.
    os.umask(0)
    os.setsid()
    _do_fork()
    sys.stdout.flush()
    sys.stderr.flush()
    sin = file(stdin, 'r')
    sout = file(stdout, 'a+')
    serr = file(stderr, 'a+', 0)
    os.dup2(sin.fileno(), sys.stdin.fileno())
    os.dup2(sout.fileno(), sys.stdout.fileno())
    os.dup2(serr.fileno(), sys.stdin.fileno())


def async_raise(tid, exctype):
    """Raise an exception within the context of a thread.

    :param tid: Thread Id.
    :param exctype: Exception class.
    :raises: exctype.
    """
    if not inspect.isclass(exctype):
        raise TypeError("Only types can be raised (not instances).")

    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(tid), ctypes.py_object(exctype)
        )

    if res == 0:
        raise ValueError("Invalid thread id.")
    elif res != 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(tid), None)
        raise SystemError("Failed to throw an exception.")

def split_dump_pattern(pattern):
    """Split a comma separated string of patterns, into a list of patterns.

    :param pattern: A comma separated string of patterns.
    """
    regex = re.compile('\s*,\s*')
    return regex.split(pattern)

def split_database_table(fully_qualified_table_name):
    """Split a fully qualified table name, which is the database name
    followed by the table name (database_name.table_name).

    :param fully_qualified_table_name: The fully qualified table name.
    """
    return fully_qualified_table_name.split('.')

def wrap_output(output):
    """Used to wrap the the output in a standard format:
    (FABRIC_UUID, VERSION_TOKEN, TTL).

    :param output: The output that needs to be wrapped.
    :return: the "output" parameter is returned in the following four
             tuple format.
    """
    return (FABRIC_UUID, VERSION_TOKEN, TTL, output)

def get_time():
    """Get current time using datetime.utcnow().
    """
    return datetime.datetime.utcnow().replace(microsecond=0)

def get_time_delta(delta):
    """Transform a value provided through the parameter delta into a
    timedelta object.

    :param delta: Delta value in seconds.
    """
    return datetime.timedelta(seconds=delta)

def get_time_from_timestamp(timestamp):
    """Return a utc time from a timestemp().
    """
    return datetime.datetime.utcfromtimestamp(timestamp).replace(microsecond=0)

def get_group_lower_bound_list(input_string):
    """Get the list of GROUP IDs and the LBs from the input string.

    :param input_string: String input by the user containing delimited
                         group ids and LBs.
    """
    group_id_list = []
    lower_bound_list = []
    group_id_lower_bound_list = input_string.replace(' ', '').split(",")
    for item in group_id_lower_bound_list:
        group_id = None
        lower_bound = None
        if item.find("/") != -1:
            group_id, lower_bound = item.split("/")
        else:
            group_id = item
        if group_id is not None:
            group_id_list.append(group_id)
        if lower_bound is not None:
            lower_bound_list.append(lower_bound)
    return group_id_list, lower_bound_list

def check_number_threads(increasing=0):
    """Check the number of threads that are running and whether the maximum
    number of connections in the state store is configured accordingly.

    :param increasing: Whether you want to increase the number of threads and
                       how many threads. Default is zero.

    It raises a ConfigurationError exception if the number of connections is
    too small.
    """
    from mysql.fabric import (
        errors as _errors,
        executor as _executor,
        persistence as _persistence,
        services as _services,
        server as _server,
    )

    n_sessions = _services.ServiceManager().get_number_sessions()
    n_executors = _executor.Executor().get_number_executors()
    n_failure_detectors = len(_server.Group.groups_by_status(_server.Group.ACTIVE))
    n_controls = 1
    persister = _persistence.current_persister()
    max_allowed_connections = persister.max_allowed_connections()
    if (n_sessions +  n_executors + n_controls + n_failure_detectors +\
        increasing) > (max_allowed_connections - 1):
        raise _errors.ConfigurationError(
            "Too many threads requested. Session threads (%s), Executor "
            "threads (%s), Control threads (%s) and Failure Detector threads "
            "(%s). The maximum number of threads allowed is (%s). Increase "
            "the maximum number of connections in the state store in order "
            "to increase this limit." % (n_sessions, n_executors, n_controls,
            n_failure_detectors, max_allowed_connections - 1)
         )



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

""" This module contains functions that are called from the services interface
and change MySQL state. Notice though that after a failure the system does no
undo the changes made through the execution of these functions.
"""
from mysql.fabric import (
    errors as _errors,
    replication as _replication,
    server as _server,
)

CONFIG_NOT_FOUND = "Configuration option not found %s . %s"

def switch_master(slave, master):
    """Make slave point to master.

    :param slave: Slave.
    :param master: Master.
    """
    _replication.stop_slave(slave, wait=True)
    _replication.switch_master(slave, master, master.user, master.passwd)
    slave.read_only = True
    _replication.start_slave(slave, wait=True)


def set_read_only(server, read_only):
    """Set server to read only mode.

    :param read_only: Either read-only or not.
    """
    server.read_only = read_only


def reset_slave(slave):
    """Stop slave and reset it.

    :param slave: slave.
    """
    _replication.stop_slave(slave, wait=True)
    _replication.reset_slave(slave, clean=True)


def process_slave_backlog(slave):
    """Wait until slave processes its backlog.

    :param slave: slave.
    """
    _replication.stop_slave(slave, wait=True)
    _replication.start_slave(slave, threads=("SQL_THREAD", ), wait=True)
    slave_status = _replication.get_slave_status(slave)[0]
    _replication.wait_for_slave(
        slave, slave_status.Master_Log_File, slave_status.Read_Master_Log_Pos
     )


def synchronize(slave, master):
    """Synchronize a slave with a master and after that stop the slave.

    :param slave: Slave.
    :param master: Master.
    """
    _replication.sync_slave_with_master(slave, master, timeout=0)


def stop_slave(slave):
    """Stop slave.

    :param slave: Slave.
    """
    _replication.stop_slave(slave, wait=True)

def read_config_value(config,  config_group,  config_name):
    """Read the value of the configuration option from the config files.

    :param config: The config class that encapsulates the config parsing
                    logic.
    :param config_group: The configuration group to which the configuration
                        belongs
    :param config_name: The name of the configuration that needs to be read,
    """
    config_value = None

    try:
        config_value =  config.get(config_group, config_name)
    except AttributeError:
        pass

    if config_value is None:
        raise _errors.ConfigurationError(CONFIG_NOT_FOUND %
                                        (config_group,  config_name))

    return config_value

def is_valid_binary(binary):
    """Prints if the binary was found in the given path.

    :param binary: The full path to the binary that needs to be verified.

    :return True: If the binary was found
        False: If the binary was not found.
    """
    import os
    return os.path.isfile(binary) and os.access(binary, os.X_OK)

def fetch_backup_server(source_group):
    """Fetch a spare, slave or master from a group in that order of
    availability. Find any spare (no criteria), if there is no spare, find
    a secondary(no criteria) and if there is no secondary the master.

    :param source_group: The group from which the server needs to
                         be fetched.
    """
    #Get a slave server whose status is spare.
    backup_server = None
    for server in source_group.servers():
        if server.status == "SPARE":
            backup_server = server

    #If there is no spare check if a running slave is available
    if backup_server is None:
        for server in source_group.servers():
            if source_group.master != server.uuid and \
                server.status == "SECONDARY":
                backup_server = server

    #If there is no running slave just use the master
    if backup_server is None:
        backup_server = _server.MySQLServer.fetch(source_group.master)

    #If there is no master throw an exception
    if backup_server is None:
        raise _errors.ShardingError(SHARD_GROUP_MASTER_NOT_FOUND)

    return backup_server



#
# Copyright (c) 2010, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This module contains and abstraction of a MySQL user object.
"""

import re

from mysql.utilities.exception import UtilError, UtilDBError, FormatError
from mysql.utilities.common.ip_parser import parse_connection, clean_IPv6


def change_user_privileges(server, user_name, user_passwd, host,
                           grant_list=None, revoke_list=None,
                           disable_binlog=False, create_user=False):
    """ Change the privileges of a new or existing user.

    This method GRANT or REVOKE privileges to a new user (creating it) or
    existing user.

    server[in]          MySQL server instances to apply changes
                        (from mysql.utilities.common.server.Server).
    user_name[in]       user name to apply changes.
    user_passwd[in]     user's password.
    host[in]            host name associated to the user account.
    grant_list[in]      List of privileges to GRANT.
    revoke_list[in]     List of privileges to REVOKE.
    disable_binlog[in]  Boolean value to determine if the binary logging
                        will be disabled to perform this operation (and
                        re-enabled at the end). By default: False (do not
                        disable binary logging).
    create_user[in]     Boolean value to determine if the user will be
                        created before changing its privileges. By default:
                        False (do no create user).
    """
    if disable_binlog:
        server.exec_query("SET SQL_LOG_BIN=0")
    if create_user:
        server.exec_query("CREATE USER '{0}'@'{1}' IDENTIFIED BY "
                          "'{2}'".format(user_name, host, user_passwd))
    if grant_list:
        grants_str = ", ".join(grant_list)
        server.exec_query("GRANT {0} ON *.* TO '{1}'@'{2}' IDENTIFIED BY "
                          "'{3}'".format(grants_str, user_name, host,
                                         user_passwd))
    if revoke_list:
        revoke_str = ", ".join(revoke_list)
        server.exec_query("REVOKE {0} ON *.* FROM '{1}'@'{2}'"
                          "".format(revoke_str, user_name, host))
    if disable_binlog:
        server.exec_query("SET SQL_LOG_BIN=1")


def parse_user_host(user_name):
    """Parse user, passwd, host, port from user:passwd@host

    user_name[in]      MySQL user string (user:passwd@host)
    """

    no_ticks = user_name.replace("'", "")
    try:
        conn_values = parse_connection(no_ticks)
    except FormatError:
        raise UtilError("Cannot parse user:pass@host : %s." %
                        no_ticks)
    return (conn_values['user'], conn_values['passwd'], conn_values['host'])


class User(object):
    """
    The User class can be used to clone the user and its grants to another
    user with the following utilities:

        - Parsing user@host:passwd strings
        - Create, Drop user
        - Check to see if user exists
        - Retrieving and printing grants for user
    """

    def __init__(self, server1, user, verbosity=0):
        """Constructor

        server1[in]        Server class
        user[in]           MySQL user credentials string (user@host:passwd)
        verbose[in]        print extra data during operations (optional)
                           default value = False
        """

        self.server1 = server1
        self.user, self.passwd, self.host = parse_user_host(user)
        self.verbosity = verbosity
        self.current_user = None
        self.query_options = {
            'fetch': False
        }

    def create(self, new_user=None):
        """Create the user

        Attempts to create the user. If the operation fails, an error is
        generated and printed.

        new_user[in]       MySQL user string (user@host:passwd)
                           (optional) If omitted, operation is performed
                           on the class instance user name.
        """

        query_str = "CREATE USER "
        user, passwd, host = None, None, None
        if new_user:
            user, passwd, host = parse_user_host(new_user)
            query_str += "'%s'@'%s' " % (user, host)
        else:
            query_str += "'%s'@'%s' " % (self.user, self.host)
            passwd = self.passwd

        if passwd:
            query_str += "IDENTIFIED BY '%s'" % (passwd)
        if self.verbosity > 0:
            print query_str

        self.server1.exec_query(query_str, self.query_options)

    def drop(self, new_user=None):
        """Drop user from the server

        Attempts to drop the user. If the operation fails, an error is
        generated and printed.

        new_user[in]       MySQL user string (user@host:passwd)
                           (optional) If omitted, operation is performed
                           on the class instance user name.
        """
        query_str = "DROP USER "
        if new_user:
            user, _, host = parse_user_host(new_user)
            query_str += "'%s'@'%s' " % (user, host)
        else:
            query_str += "'%s'@'%s' " % (self.user, self.host)

        if self.verbosity > 0:
            print query_str

        try:
            self.server1.exec_query(query_str, self.query_options)
        except UtilError:
            return False
        return True

    def exists(self, user_name=None):
        """Check to see if the user exists

        user_name[in]      MySQL user string (user@host:passwd)
                           (optional) If omitted, operation is performed
                           on the class instance user name.

        return True = user exists, False = user does not exist
        """

        user, host, _ = self.user, self.host, self.passwd
        if user_name:
            user, _, host = parse_user_host(user_name)

        res = self.server1.exec_query("SELECT * FROM mysql.user "
                                      "WHERE user = %s and host = %s",
                                      {'params': (user, host)})

        return (res is not None and len(res) >= 1)

    def get_grants(self, globals_privs=False):
        """Retrieve the grants for the current user

        globals_privs[in]     Include global privileges in clone (i.e. user@%)

        returns result set or None if no grants defined
        """
        # Get the users' connection user@host if not retrieved
        if self.current_user is None:
            res = self.server1.exec_query("SELECT CURRENT_USER()")
            parts = res[0][0].split('@')
            # If we're connected as some other user, use the user@host
            # defined at instantiation
            if parts[0] != self.user:
                host = clean_IPv6(self.host)
                self.current_user = "'%s'@'%s'" % (self.user, host)
            else:
                self.current_user = "'%s'@'%s'" % (parts[0], parts[1])
        grants = []
        try:
            res = self.server1.exec_query("SHOW GRANTS FOR "
                                          "{0}".format(self.current_user))
            for grant in res:
                grants.append(grant)
        except UtilDBError:
            pass  # Error here is ok - no grants found.
        if globals_privs:
            try:
                res = self.server1.exec_query("SHOW GRANTS FOR "
                                              "'{0}'{1}".format(self.user,
                                                                "@'%'"))
                for grant in res:
                    grants.append(grant)
            except UtilDBError:
                pass  # Error here is ok - no grants found.
        return grants

    def has_privilege(self, db, obj, access, allow_skip_grant_tables=True):
        """Check to see user has a specific access to a db.object.

        db[in]             Name of database
        obj[in]            Name of object
        access[in]         MySQL privilege to check (e.g. SELECT, SUPER, DROP)
        allow_skip_grant_tables[in]  If True, allow silent failure for
                           cases where the server is started with
                           --skip-grant-tables. Default=True

        Returns True if user has access, False if not
        """
        grants_enabled = self.server1.grant_tables_enabled()
        # If grants are disabled and it is Ok to allow skipped grant tables,
        # return True - privileges disabled so user can do anything.
        if allow_skip_grant_tables and not grants_enabled:
            return True
        # Convert privilege to upper cases.
        access = access.upper()
        # Create regexp to match SHOW GRANTS.
        if access == "GRANT OPTION":
            # WITH GRANT OPTION appears at the end of SHOW GRANTS.
            regex = re.compile(r"GRANT.+ON\s+"
                               r"(?:\*|['`]?{db}['`]?)\.(?:\*|[`']?{obj}[`']?)"
                               r"\s+TO.+"
                               r"WITH GRANT OPTION".format(db=re.escape(db),
                                                           obj=re.escape(obj)))
        else:
            # GRANT with ALL PRIVILEGES or given privilege (access parameter).
            regex = re.compile(r"GRANT.*\b(?:ALL PRIVILEGES|{priv})\b.*ON\s+"
                               r"(?:\*|['`]?{db}['`]?)\.(?:\*|[`']?{obj}[`']?)"
                               r"\s+TO".format(priv=re.escape(access),
                                               db=re.escape(db),
                                               obj=re.escape(obj)))
        for grant in self.get_grants(True):
            # Check if SHOW GRANTS match regexp with privilege.
            if regex.match(grant[0]):
                return True

    def print_grants(self):
        """Display grants for the current user"""

        res = self.get_grants(True)
        for grant_tuple in res:
            print grant_tuple[0]

    def clone(self, new_user, destination=None, globals_privs=False):
        """Clone the current user to the new user

        Operation will create the new user account copying all of the
        grants for the current user to the new user. If operation fails,
        an error message is generated and the process halts.

        new_name[in]       MySQL user string (user@host:passwd)
        destination[in]    A connection to a new server to clone the user
                           (default is None)
        globals_privs[in]  Include global privileges in clone (i.e. user@%)

        Note: Caller must ensure the new user account does not exist.
        """

        res = self.get_grants(globals_privs)
        server = self.server1
        if destination is not None:
            server = destination
        for row in res:
            # Create an instance of the user class.
            user = User(server, new_user, self.verbosity)
            if not user.exists():
                user.create()

            base_user_ticks = "'" + self.user + "'@'" + self.host + "'"
            user, _, host = parse_user_host(new_user)
            new_user_ticks = "'" + user + "'@'" + host + "'"
            grant = row[0].replace(base_user_ticks, new_user_ticks, 1)

            # Need to remove the IDENTIFIED BY clause for the base user.
            search_str = "IDENTIFIED BY PASSWORD"
            try:
                start = grant.index(search_str)
            except:
                start = 0

            if start > 0:
                end = grant.index("'", start + len(search_str) + 2) + 2
                grant = grant[0:start] + grant[end:]

            if self.verbosity > 0:
                print grant

            res = server.exec_query(grant, self.query_options)



#!/usr/bin/env python
#
# Copyright (c) 2013, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This file contains the .frm file utility for reading .frm files to construct
CREATE TABLE commands and display diagnostic information.
"""

import os
import sys

from mysql.utilities.common.tools import (check_python_version,
                                          check_connector_python)

# Check Python version compatibility
check_python_version()

# Check C/Python version compatibility
if not check_connector_python():
    sys.exit(1)

from mysql.utilities import VERSION_FRM
from mysql.utilities.exception import FormatError, UtilError
from mysql.utilities.command.read_frm import (read_frm_files,
                                              read_frm_files_diagnostic)
from mysql.utilities.common.options import CaseInsensitiveChoicesOption
from mysql.utilities.common.ip_parser import parse_connection
from mysql.utilities.common.options import (add_verbosity, license_callback,
                                            UtilitiesParser)
from mysql.utilities.common.server import connect_servers


class MyParser(UtilitiesParser):
    """Custom class to set the epilog.
    """
    def format_epilog(self, formatter):
        return self.epilog

# Constants
NAME = "MySQL Utilities - mysqlfrm "
DESCRIPTION = "mysqlfrm - show CREATE TABLE from .frm files"
USAGE = ("%prog --server=[user[:<pass>]@host[:<port>][:<socket>]|"
         "<login-path>[:<port>][:<socket>]] [path\\tbl1.frm|db:tbl.frm]")
EXTENDED_HELP = """
Introduction
------------
The mysqlfrm utility is designed as a recovery tool that reads .frm files and
produces facsimile CREATE statements from the table definition data found in
the .frm file. In most cases, the CREATE statement produced will be usable
for recreating the table on another server or for extended diagnostics.
However, some features are not saved in the .frm files and therefore will be
omitted. The exclusions include but are not limited to:

  - foreign key constraints
  - auto increment number sequences

The mysqlfrm utility has two modes of operation. The default mode is
designed to spawn an instance of an installed server by reference
to the base directory using the --basedir option or by connecting to the
server with the --server option. The process will not alter the original
.frm file(s). This mode also requires the --port option to specify a
port to use for the spawned server. The spawned server will be shutdown
and all temporary files removed after the .frm files are read.

A diagnostic mode is available by using the --diagnostic option. This will
switch the utility to reading the .frm files byte-by-byte to recover as
much information as possible. The diagnostic mode has additional limitations
in that it cannot decipher character set or collation values without using
an existing server installation specified with either the --server or
--basedir option. This can also affect the size of the columns if the table
uses multi-byte characters. Use this mode when the default mode cannot read
the file or if there is no server installed on the host.

To read .frm files, list each file as a separate argument for the utility as
shown in the following examples. You will need to specify the path for each
.frm file you want to read or supply a path to a directory and all of the
.frm files in that directory will be read.

  # Read a single .frm file in the default mode using the server installed
  # in /usr/local/bin/mysql where the .frm file is in the current folder.
  # Notice the use of the db:table.frm format for specifying the database
  # name for the table. The database name appears to the left of ':' and
  # the .frm name to the right. So in this case, we have database = test1
  # and table = db1 so the CREATE statement will read CREATE test1.db1.

  $ mysqlfrm --basedir=/usr/local/bin/mysql test1:db1.frm --port=3333

  # Read multiple .frm files in the default mode using a running server
  # where the .frm files are located in different folders.

  $ mysqlfrm --server=root:pass@localhost:3306 /mysql/data/temp1/t1.frm \\
             /mysql/data/temp2/g1.frm --port=3310

  # Execute the spawned server under a different user name and read
  # all of the .frm files in a particular folder in default mode.

  $ mysqlfrm --server=root:pass@localhost:3306 /mysql/data/temp1/t1.frm \\
             /mysql/data/temp2/g1.frm --port=3310 --user=joeuser

  # Read all of the .frm files in a particular folder using the diagnostic
  # mode.

  $ mysqlfrm --diagnostic /mysql/data/database1



Helpful Hints
-------------
  - Tables with certain storage engines cannot be read in the default mode.
    These include PARTITION, PERFORMANCE_SCHEMA. You must read these with
    the --diagnostic mode.

  - Use the --diagnostic mode for tables that fail to open correctly
    in the default mode or if there is no server installed on the host.

  - To change the storage engine in the CREATE statement generated for all
    .frm files read, use the --new-storage-engine option

  - To turn off all messages except the CREATE statement and warnings or
    errors, use the --quiet option.

  - Use the --show-stats option to see file statistics for each .frm file.

  - If you encounter connection or similar errors when running in default
    mode, re-run the command with the --verbosity option and view the
    output from the spawned server and repair any errors in launching the
    server. If mysqlfrm fails in the middle, you may need to manually
    shutdown the server on the port specified with --port.

  - If the spawned server takes more than 10 seconds to start, use the
    --start-timeout option to increase the timeout to wait for the
    spawned server to start.

  - If you need to run the utility with elevated privileges, use the --user
    option to execute the spawned server using a normal user account.

  - You can specify the database name to be used in the resulting CREATE
    statement by prepending the .frm file with the name of the database
    followed by a colon. For example, oltp:t1.frm will use 'oltp' for the
    database name in the CREATE statement. The optional database name can
    also be used with paths. For example, /home/me/oltp:t1.frm will use
    'oltp' as the database name. If you leave off the optional database
    name and include a path, the last folder will be the database name.
    For example /home/me/data1/t1.frm will use 'data1' as the database
    name. If you do not want to use the last folder as the database name,
    simply specify the colon like this: /home/me/data1/:t1.frm. In this
    case, the database will be omitted from the CREATE statement.

  - If you use the --new-storage-engine option, you must also provide the
    --frmdir option. When these options are specified, the utility will
    generate a new .frm file (prefixed with 'new_') and save it in the
    --frmdir= directory.

Enjoy!

"""

if __name__ == '__main__':
    # Setup the command parser
    program = os.path.basename(sys.argv[0]).replace(".py", "")
    parser = MyParser(
        version=VERSION_FRM.format(program=program),
        description=DESCRIPTION,
        usage=USAGE,
        add_help_option=False,
        option_class=CaseInsensitiveChoicesOption,
        epilog=EXTENDED_HELP,
        prog=program
    )

    # Add --License option
    parser.add_option("--license", action='callback',
                      callback=license_callback,
                      help="display program's license and exit")

    # Add --help option
    parser.add_option("--help", action="help")

    # Setup utility-specific options:

    # Add --basedir option
    parser.add_option("--basedir", action="store", dest="basedir",
                      default=None, type="string",
                      help="the base directory for the server")

    # Add diagnostic mode
    parser.add_option("--diagnostic", action="store_true", dest="diagnostic",
                      help="read the frm files byte-by-byte to form the "
                           "CREATE statement. May require the --server or "
                           "--basedir options to decipher character set "
                           "information")
    # Add engine
    parser.add_option("--new-storage-engine", action="store",
                      dest="new_engine", default=None,
                      help="change ENGINE clause to use this engine.")

    # Add frmdir
    parser.add_option("--frmdir", action="store", dest="frmdir", default=None,
                      help="save the new .frm files in this directory. "
                           "Used and valid with --new-storage-engine only.")

    # Need port - only valid with --diagnostic mode
    parser.add_option("--port", action="store", dest="port",
                      help="Port to use for the spawned server.", default=None)

    # Add show-stats
    parser.add_option("--show-stats", "-s", action="store_true",
                      dest="show_stats",
                      help="show file statistics and general table "
                           "information.")

    # Add server option
    parser.add_option("--server", action="store", dest="server", type="string",
                      default=None,
                      help="connection information for the server in the "
                           "form: <user>[:<password>]@<host>[:<port>]"
                           "[:<socket>] or <login-path>[:<port>][:<socket>] "
                           "(optional) - if provided, the storage engine and "
                           "character set information will be validated "
                           "against this server.")

    # Add user option
    parser.add_option("--user", action="store", dest="user", type="string",
                      default=None,
                      help="user account to launch spawned server. Required "
                           "if running as root user. Used only in the "
                           "default mode.")

    # Add startup timeout
    parser.add_option("--start-timeout", action="store", dest="start_timeout",
                      type=int, default=10,
                      help="Number of seconds to wait for spawned server to "
                           "start. Default = 10.")

    # Add verbosity mode
    add_verbosity(parser, True)

    # Now we process the rest of the arguments.
    opt, args = parser.parse_args()

    # Check for argument errors
    if not args:
        parser.error("Nothing to do. You must specify a list of paths or "
                     "files to read. See --help for more information and "
                     "examples.")

    if not opt.port and not opt.diagnostic:
        parser.error("The --port option is required for reading .frm files in "
                     "the default mode.")

    if opt.diagnostic and opt.port:
        print("# WARNING The --port option is not used in the "
              "--diagnostic mode.")

    use_port = opt.port
    if not opt.diagnostic and opt.port:
        try:
            use_port = int(opt.port)
        except ValueError:
            parser.error("The --port option requires an integer value.")

    # Check for access to basedir if specified
    if opt.basedir:
        opt.basedir = os.path.expanduser(opt.basedir)
        if not os.access(opt.basedir, os.R_OK):
            parser.error("You must have read access to the base directory "
                         "specified with the --basedir option.")

    # Warn if both --server and --basedir used
    if opt.server and opt.basedir:
        print ("# WARNING: The --server option is not needed when the "
               "--basedir option is used.")

    # Warn if --diagnostic and --user
    if opt.diagnostic and opt.user:
        print ("# WARNING: The --user option is only used for the default "
               "mode.")

    # Check for --new-storage-engine and --frmdir
    if opt.new_engine:
        if not opt.frmdir:
            parser.error("You must specify the --frmdir with "
                         "--new-storage-engine.")
        # Check frmdir validity
        else:
            if not os.path.exists(opt.frmdir):
                parser.error("The directory, "
                             "'{0}' does not exist.".format(opt.frmdir))
            if not os.access(opt.frmdir, os.R_OK | os.W_OK):
                parser.error("You must have read and write access to the .frm "
                             "directory '{0}'.".format(opt.frmdir))
    elif not opt.new_engine and opt.frmdir:
        print("# WARNING: --frmdir encountered without --new-storage-engine. "
              "No .frm files will be saved.")

    server = None
    if opt.server is None and opt.diagnostic:
        print("# WARNING: Cannot generate character set or "
              "collation names without the --server option.")

    # Check start timeout for minimal value
    if int(opt.start_timeout) < 10:
        opt.start_timeout = 10
        print("# WARNING: --start-timeout must be >= 10 seconds. Using "
              "default value.")

    # Parse source connection values if --server provided
    if opt.server is not None and not opt.basedir:
        try:
            source_values = parse_connection(opt.server, None, {})
        except FormatError as err:
            parser.error("Source connection values invalid: %s." % err)
        except UtilError as err:
            parser.error("Source connection values invalid: %s." % err.errmsg)

        try:
            conn_options = {
                'version': "5.1.30",
                'quiet': opt.quiet,
            }
            servers = connect_servers(source_values, None, conn_options)
        except UtilError as error:
            parser.error(error.errmsg)
        server = servers[0]

        if use_port == int(server.port):
            parser.error("You must specify a different port to use for the "
                         "spawned server.")

        basedir = server.show_server_variable("basedir")[0][1]
    else:
        basedir = opt.basedir

    # Set options for frm operations.
    options = {
        "basedir": basedir,
        "new_engine": opt.new_engine,
        "show_stats": opt.show_stats,
        "port": use_port,
        "quiet": opt.quiet,
        "server": server,
        "verbosity": opt.verbosity if opt.verbosity else 0,
        "user": opt.user,
        "start_timeout": opt.start_timeout,
        "frm_dir": opt.frmdir,
    }

    # Print disclaimer banner for diagnostic mode
    if opt.diagnostic:
        print ("# CAUTION: The diagnostic mode is a best-effort parse of the "
               ".frm file. As such, it may not identify all of the components "
               "of the table correctly. This is especially true for damaged "
               "files. It will also not read the default values for the "
               "columns and the resulting statement may not be syntactically "
               "correct.")

    if os.name == "posix":
        if os.getuid() == 0 and not opt.diagnostic and not opt.user:
            parser.error("Running a spawned server as root is not advised. If "
                         "you want to run the utility as root, please provide "
                         "the --user option to specify a user to use to "
                         "launch the server. Example: --user=mysql.")

    all_frm_files = []
    for arg in args:
        frm_files_found = []

        # check to see if we have access iff it is not in the form of
        # db:table.frm, but watchout for Windows paths!
        if (os.name != "nt" and ":" not in arg) or \
           (os.name == "nt" and len(arg) >= 2 and ":" not in arg[2:]):
            if not os.access(arg, os.R_OK):
                print ("ERROR: Cannot read %s. You must have read privileges"
                       " to the file or path and it must exist. Skipping "
                       "this argument." % arg)
                continue

        # if argument is a folder, get all files from the folder and read them
        if os.path.isdir(arg):
            files = os.listdir(arg)
            for filename in files:
                if os.path.splitext(filename)[1].lower() == ".frm":
                    frm_files_found.append(os.path.join(arg, filename))
        else:
            frm_files_found.append(arg)

        if not frm_files_found:
            print "# NOTE: No .frm files found in folder %s." % arg
            continue
        all_frm_files.extend(frm_files_found)

    # For each file specified, attempt to read the .frm file
    try:
        all_frm_files.sort()
        if opt.diagnostic:
            read_frm_files_diagnostic(all_frm_files, options)
        else:
            failed = read_frm_files(all_frm_files, options)
            if failed:
                print ("#\n# WARNING: The following files could not be read. "
                       "You can try the --diagnostic mode to read these "
                       "files.\n#")
                for frm_file in failed:
                    print "#", frm_file
                print "#"
    except UtilError as error:
        print "ERROR: {0}".format(error.errmsg)
        sys.exit(1)

    if not opt.quiet:
        print "#...done."

    sys.exit()



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#
"""This modules contains a simple failure detector which is used by Fabric
to monitor the availability of servers within groups.

If a master cannot be accessed through the
:meth:`~mysql.fabric.server.MySQLServer.is_alive` method after `n` consecutive
attempts, the failure detector considers that it has failed and proceeds with
the election of a new master. The failure detector does not choose any new
master but only triggers the :const:`~mysql.fabric.events.REPORT_FAILURE` event
which responsible for doing so.

If a slave cannot be accessed either the same event is triggered but in this
case the server is only marked as faulty.

See :meth:`~mysql.fabric.server.MySQLServer.is_alive`.
See :class:`~mysql.fabric.services.highavailability.PromoteMaster`.
See :class:`~mysql.fabric.services.servers.ReportFailure`.
"""
import threading
import time
import logging

from mysql.fabric import (
    errors as _errors,
    persistence as _persistence,
    config as _config,
    executor as _executor,
)

from mysql.fabric.events import (
    trigger,
)

from mysql.fabric.utils import (
    get_time,
)

_LOGGER = logging.getLogger(__name__)

class FailureDetector(object):
    """Responsible for periodically checking if a set of servers within a
    group is alive.

    It does so by connecting to these servers and executing a query (i.e.
    :meth:`~mysql.fabric.server.MySQLServer.is_alive`.
    """
    LOCK = threading.Condition()
    GROUPS = {}

    _MIN_DETECTION_INTERVAL = 2.0
    _DETECTION_INTERVAL = _DEFAULT_DETECTION_INTERVAL = 5.0

    _MIN_DETECTIONS = 1
    _DETECTIONS = _DEFAULT_DETECTIONS = 2

    _MIN_DETECTION_TIMEOUT = 1
    _DETECTION_TIMEOUT = _DEFAULT_DETECTION_TIMEOUT = 1

    @staticmethod
    def register_groups():
        """Upon startup initializes a failure detector for each group.
        """
        from mysql.fabric.server import Group
        _LOGGER.info("Starting failure detector.")
        for row in Group.groups_by_status(Group.ACTIVE):
            FailureDetector.register_group(row[0])

    @staticmethod
    def register_group(group_id):
        """Start a failure detector for a group.

        :param group_id: Group's id.
        """
        _LOGGER.info("Monitoring group (%s).", group_id)
        with FailureDetector.LOCK:
            if group_id not in FailureDetector.GROUPS:
                detector = FailureDetector(group_id)
                detector.start()
                FailureDetector.GROUPS[group_id] = detector

    @staticmethod
    def unregister_group(group_id):
        """Stop a failure detector for a group.

        :param group_id: Group's id.
        """
        _LOGGER.info("Stop monitoring group (%s).", group_id)
        with FailureDetector.LOCK:
            if group_id in FailureDetector.GROUPS:
                detector = FailureDetector.GROUPS[group_id]
                detector.shutdown()
                del FailureDetector.GROUPS[group_id]

    @staticmethod
    def unregister_groups():
        """Upon shutdown stop all failure detectors that are running.
        """
        _LOGGER.info("Stopping failure detector.")
        with FailureDetector.LOCK:
            for detector in FailureDetector.GROUPS.values():
                detector.shutdown()
            FailureDetector.GROUPS = {}

    def __init__(self, group_id):
        """Constructor for FailureDetector.
        """
        self.__group_id = group_id
        self.__thread = None
        self.__check = False

    def start(self):
        """Start the failure detector.
        """
        self.__check = True
        self.__thread = threading.Thread(target=self._run,
            name="FailureDetector(" + self.__group_id + ")")
        self.__thread.daemon = True
        self.__thread.start()

    def shutdown(self):
        """Stop the failure detector.
        """
        self.__check = False

    def _run(self):
        """Function that verifies servers' availabilities.
        """
        from mysql.fabric.server import (
            Group,
            MySQLServer,
        )

        ignored_status = [MySQLServer.FAULTY]
        quarantine = {}
        interval = FailureDetector._DETECTION_INTERVAL
        detections = FailureDetector._DETECTIONS
        detection_timeout = FailureDetector._DETECTION_TIMEOUT

        _persistence.init_thread()

        while self.__check:
            try:
                unreachable = set()
                group = Group.fetch(self.__group_id)
                if group is not None:
                    for server in group.servers():
                        if server.status in ignored_status or \
                            server.is_alive(detection_timeout):
                            continue

                        unreachable.add(server.uuid)

                        _LOGGER.warning(
                            "Server (%s) in group (%s) is unreachable.",
                            server.uuid, self.__group_id
                        )

                        unstable = False
                        failed_attempts = 0
                        if server.uuid not in quarantine:
                            quarantine[server.uuid] = failed_attempts = 1
                        else:
                            failed_attempts = quarantine[server.uuid] + 1
                            quarantine[server.uuid] = failed_attempts
                        if failed_attempts >= detections:
                            unstable = True

                        can_set_faulty = group.can_set_server_faulty(
                            server, get_time()
                        )
                        if unstable and can_set_faulty:
                            procedures = trigger("REPORT_FAILURE",
                                set([self.__group_id]), str(server.uuid),
                                threading.current_thread().name,
                                MySQLServer.FAULTY, False
                            )
                            executor = _executor.Executor()
                            for procedure in procedures:
                                executor.wait_for_procedure(procedure)

                for uuid in quarantine.keys():
                    if uuid not in unreachable:
                        del quarantine[uuid]

            except (_errors.ExecutorError, _errors.DatabaseError):
                pass
            except Exception as error:
                _LOGGER.exception(error)

            time.sleep(interval / detections)

        _persistence.deinit_thread()


def configure(config):
    """Set configuration values.
    """
    try:
        detection_interval = \
            float(config.get("failure_tracking", "detection_interval"))
        if detection_interval < FailureDetector._MIN_DETECTION_INTERVAL:
            _LOGGER.warning(
                "Detection interval cannot be lower than %s.",
                FailureDetector._MIN_DETECTION_INTERVAL
            )
            detection_interval = FailureDetector._MIN_DETECTION_INTERVAL
        FailureDetector._DETECTION_INTERVAL = float(detection_interval)
    except (_config.NoOptionError, _config.NoSectionError, ValueError):
        pass

    try:
        detections = int(config.get("failure_tracking", "detections"))
        if detections < FailureDetector._MIN_DETECTIONS:
            _LOGGER.warning(
                "Detections cannot be lower than %s.",
                FailureDetector._MIN_DETECTIONS
            )
            detections = FailureDetector._MIN_DETECTIONS
        FailureDetector._DETECTIONS = int(detections)
    except (_config.NoOptionError, _config.NoSectionError, ValueError):
        pass

    try:
        detection_timeout = \
            int(config.get("failure_tracking", "detection_timeout"))
        if detection_timeout < FailureDetector._MIN_DETECTION_TIMEOUT:
            _LOGGER.warning(
                "Detection timeout cannot be lower than %s.",
                FailureDetector._MIN_DETECTION_TIMEOUT
            )
            detection_interval = FailureDetector._MIN_DETECTION_TIMEOUT
        FailureDetector._DETECTION_TIMEOUT = int(detection_timeout)
    except (_config.NoOptionError, _config.NoSectionError, ValueError):
        pass
